# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from spdx_tools.spdx3.model.software.software_purpose import SoftwarePurpose
from spdx_tools.spdx3.model.software.file import File
from spdx_tools.spdx3.model.software.package import Package
from spdx_tools.spdx3.model.software.snippet import Snippet
from spdx_tools.spdx3.model.software.sbom import Sbom, SBOMType
from spdx_tools.spdx3.model.software.software_dependency_relationship import (
    SoftwareDependencyRelationship,
    SoftwareDependencyLinkType,
    DependencyConditionalityType,
)



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import Any, Type

from spdx_tools.spdx.jsonschema.converter import TypedConverter
from spdx_tools.spdx.jsonschema.json_property import JsonProperty
from spdx_tools.spdx.jsonschema.package_verification_code_properties import PackageVerificationCodeProperty
from spdx_tools.spdx.model import Document, PackageVerificationCode


class PackageVerificationCodeConverter(TypedConverter[PackageVerificationCode]):
    def _get_property_value(
        self,
        verification_code: PackageVerificationCode,
        verification_code_property: PackageVerificationCodeProperty,
        document: Document = None,
    ) -> Any:
        if verification_code_property == PackageVerificationCodeProperty.PACKAGE_VERIFICATION_CODE_EXCLUDED_FILES:
            return verification_code.excluded_files or None
        elif verification_code_property == PackageVerificationCodeProperty.PACKAGE_VERIFICATION_CODE_VALUE:
            return verification_code.value

    def get_json_type(self) -> Type[JsonProperty]:
        return PackageVerificationCodeProperty

    def get_data_model_type(self) -> Type[PackageVerificationCode]:
        return PackageVerificationCode



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

from typing import List

from spdx_tools.spdx.validation.external_document_ref_validator import validate_external_document_ref
from spdx_tools.spdx.validation.validation_message import ValidationMessage
from tests.spdx.fixtures import external_document_ref_fixture


def test_valid_external_document_ref():
    external_document_ref = external_document_ref_fixture()
    validation_messages: List[ValidationMessage] = validate_external_document_ref(
        external_document_ref, "parent_id", "SPDX-2.3"
    )

    assert validation_messages == []



# SPDX-License-Identifier: Apache-2.0
#  Copyright (c) 2022 spdx contributors
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#    http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from beartype.typing import TextIO

from spdx_tools.spdx.model import Document, Relationship, RelationshipType
from spdx_tools.spdx.writer.tagvalue.annotation_writer import write_annotation
from spdx_tools.spdx.writer.tagvalue.creation_info_writer import write_creation_info
from spdx_tools.spdx.writer.tagvalue.extracted_licensing_info_writer import write_extracted_licensing_info
from spdx_tools.spdx.writer.tagvalue.file_writer import write_file
from spdx_tools.spdx.writer.tagvalue.package_writer import write_package
from spdx_tools.spdx.writer.tagvalue.relationship_writer import write_relationship
from spdx_tools.spdx.writer.tagvalue.snippet_writer import write_snippet
from spdx_tools.spdx.writer.tagvalue.tagvalue_writer_helper_functions import (
    get_file_ids_with_contained_snippets,
    scan_relationships,
    write_list_of_elements,
    write_optional_heading,
    write_separator,
)
from spdx_tools.spdx.writer.write_utils import validate_and_deduplicate


def write_document_to_stream(document: Document, stream: TextIO, validate: bool = True, drop_duplicates: bool = True):
    document = validate_and_deduplicate(document, validate, drop_duplicates)
    write_document(document, stream)


def write_document_to_file(document: Document, file_name: str, validate: bool = True, drop_duplicates: bool = True):
    with open(file_name, "w") as out:
        write_document_to_stream(document, out, validate, drop_duplicates)


def write_document(document: Document, text_output: TextIO):
    relationships_to_write, contained_files_by_package_id = scan_relationships(
        document.relationships, document.packages, document.files
    )
    file_ids_with_contained_snippets = get_file_ids_with_contained_snippets(document.snippets, document.files)
    packaged_file_ids = [file.spdx_id for files_list in contained_files_by_package_id.values() for file in files_list]
    filed_snippet_ids = [
        snippet.spdx_id for snippets_list in file_ids_with_contained_snippets.values() for snippet in snippets_list
    ]

    text_output.write("## Document Information\n")
    write_creation_info(document.creation_info, text_output)
    write_separator(text_output)

    for snippet in document.snippets:
        if snippet.spdx_id not in filed_snippet_ids:
            write_snippet(snippet, text_output)
            write_separator(text_output)

    for file in document.files:
        if file.spdx_id not in packaged_file_ids:
            write_file(file, text_output)
            write_separator(text_output)
            if file.spdx_id in file_ids_with_contained_snippets:
                write_list_of_elements(
                    file_ids_with_contained_snippets[file.spdx_id], write_snippet, text_output, with_separator=True
                )

    already_written_file_ids = []  # a file can belong to multiple packages but must appear only once
    for package in document.packages:
        write_package(package, text_output)
        write_separator(text_output)
        if package.spdx_id in contained_files_by_package_id:
            for file in contained_files_by_package_id[package.spdx_id]:
                if file.spdx_id in already_written_file_ids:
                    relationships_to_write.append(
                        Relationship(package.spdx_id, RelationshipType.CONTAINS, file.spdx_id)
                    )
                else:
                    write_file(file, text_output)
                    write_separator(text_output)
                    if file.spdx_id in file_ids_with_contained_snippets:
                        write_list_of_elements(
                            file_ids_with_contained_snippets[file.spdx_id],
                            write_snippet,
                            text_output,
                            with_separator=True,
                        )
                    already_written_file_ids.append(file.spdx_id)

    write_optional_heading(document.extracted_licensing_info, "## License Information\n", text_output)
    write_list_of_elements(
        document.extracted_licensing_info, write_extracted_licensing_info, text_output, with_separator=True
    )

    write_optional_heading(relationships_to_write, "## Relationships\n", text_output)
    write_list_of_elements(relationships_to_write, write_relationship, text_output)
    write_separator(text_output)

    write_optional_heading(document.annotations, "## Annotations\n", text_output)
    write_list_of_elements(document.annotations, write_annotation, text_output, with_separator=True)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import json
import os.path

# current workflow: markdown files + spec_parser -> model.ttl -> convert to json_ld: SPDX_OWL.json ->
# use the function below to generate context.json
# properties with Enum range should look like this (probably), so that their values are automatically appended
# with the Enum URI:
# "annotationType": {
#     "@id": "core:annotationType",
#     "@type": "@vocab",
#     "@context": {
#         "@vocab": "core:AnnotationType/"
#     }
# },

PROPERTIES_WITH_ENUM_RANGE = [
    "safetyRiskAssessment",
    "sensitivePersonalInformation",
    "annotationType",
    "externalIdentifierType",
    "externalReferenceType",
    "algorithm",
    "scope",
    "profile",
    "completeness",
    "relationshipType",
    "confidentialityLevel",
    "datasetAvailability",
    "decisionType",
    "justificationType",
    "catalogType",
    "conditionality",
    "sbomType",
    "softwareLinkage",
    "purpose",
]

REFERENCE_PROPERTY_TYPES = [
    "core:Element",
    "core:Agent",
]


def convert_spdx_owl_to_jsonld_context(spdx_owl: str = "SPDX_OWL.json"):
    with open(spdx_owl, "r") as infile:
        owl_dict = json.load(infile)

    context_dict = owl_dict["@context"]

    for node in owl_dict["@graph"]:
        # print(node)
        node_type = node.get("@type")
        if not node_type:
            # print(node)
            continue

        if "owl:NamedIndividual" in node_type:
            continue
        elif node_type in ["owl:DatatypeProperty", "owl:ObjectProperty"]:
            name = node["@id"].split(":")[-1]
            type_id = node["rdfs:range"]["@id"]

            if name in context_dict and context_dict[name]["@id"].startswith("core"):
                # if in doubt, prioritize core properties
                continue

            if name in PROPERTIES_WITH_ENUM_RANGE:
                if name == "profile":
                    # FIXME: since the allowed values for the profile enum collide with
                    # our namespaces, we need to explicitly remap their meaning in the context
                    context_dict[name] = {
                        "@id": node["@id"],
                        "@type": "@vocab",
                        "@context": {
                            "core": "https://spdx.org/rdf/Core/ProfileIdentifierType/core",
                            "software": "https://spdx.org/rdf/Core/ProfileIdentifierType/software",
                            "licensing": "https://spdx.org/rdf/Core/ProfileIdentifierType/licensing",
                            "security": "https://spdx.org/rdf/Core/ProfileIdentifierType/security",
                            "build": "https://spdx.org/rdf/Core/ProfileIdentifierType/build",
                            "ai": "https://spdx.org/rdf/Core/ProfileIdentifierType/ai",
                            "dataset": "https://spdx.org/rdf/Core/ProfileIdentifierType/dataset",
                            "usage": "https://spdx.org/rdf/Core/ProfileIdentifierType/usage",
                            "extension": "https://spdx.org/rdf/Core/ProfileIdentifierType/extension",
                        },
                    }
                else:
                    context_dict[name] = {
                        "@id": node["@id"],
                        "@type": "@vocab",
                        "@context": {"@vocab": type_id + "/"},
                    }
            elif node_type == "owl:ObjectProperty" and type_id in REFERENCE_PROPERTY_TYPES:
                context_dict[name] = {"@id": node["@id"], "@type": "@id"}
            else:
                context_dict[name] = {"@id": node["@id"], "@type": type_id}

        elif node_type == "owl:Class":
            name = node["@id"].split(":")[-1]
            context_dict[name] = node["@id"]

        elif isinstance(node_type, list):
            name = node["@id"].split(":")[-1]
            context_dict[name] = node["@id"]

        else:
            print(f"unknown node_type: {node_type}")

    with open(os.path.join(os.path.dirname(__file__), "context.json"), "w") as infile:
        json.dump(context_dict, infile)


if __name__ == "__main__":
    convert_spdx_owl_to_jsonld_context("SPDX_OWL.json")



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from typing import Optional

from beartype.typing import List

from spdx_tools.spdx3.model import CreationInfo, ExternalIdentifier, ExternalIdentifierType, Organization, Person, Tool
from spdx_tools.spdx3.payload import Payload
from spdx_tools.spdx.model.actor import Actor as Spdx2_Actor
from spdx_tools.spdx.model.actor import ActorType


def bump_actor(
    spdx2_actor: Spdx2_Actor, payload: Payload, document_namespace: str, creation_info: Optional[CreationInfo] = None
) -> str:
    name: str = spdx2_actor.name
    email: str = spdx2_actor.email
    actor_type: ActorType = spdx2_actor.actor_type

    external_identifiers: List[ExternalIdentifier] = []
    name_without_whitespace = "".join(name.split())
    if email:
        external_identifiers.append(ExternalIdentifier(ExternalIdentifierType.EMAIL, email))
        spdx_id: str = f"{document_namespace}#SPDXRef-Actor-{name_without_whitespace}-{email}"
    else:
        spdx_id: str = f"{document_namespace}#SPDXRef-Actor-{name_without_whitespace}"

    if spdx_id in payload.get_full_map():  # the agent/tool already exists, so we don't need to create a new one
        return spdx_id

    value_dict = {
        "spdx_id": spdx_id,
        "creation_info": creation_info,
        "name": name,
        "external_identifier": external_identifiers,
    }

    if actor_type == ActorType.PERSON:
        agent_or_tool = Person(**value_dict)

    elif actor_type == ActorType.ORGANIZATION:
        agent_or_tool = Organization(**value_dict)

    elif actor_type == ActorType.TOOL:
        agent_or_tool = Tool(**value_dict)

    else:
        raise ValueError(f"no conversion rule defined for ActorType {actor_type}")

    payload.add_element(agent_or_tool)

    return spdx_id



# SPDX-License-Identifier: Apache-2.0
#  Copyright (c) 2023 spdx contributors
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from beartype.typing import TextIO

from spdx_tools.spdx3.model import Tool
from spdx_tools.spdx3.writer.console.element_writer import write_element_properties


def write_tool(tool: Tool, text_output: TextIO, heading: bool = True):
    if heading:
        text_output.write("## Tool\n")
    write_element_properties(tool, text_output)



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

import re

from beartype.typing import List
from uritools import isabsuri, urisplit

url_pattern = (
    "(http:\\/\\/www\\.|https:\\/\\/www\\.|http:\\/\\/|https:\\/\\/|ssh:\\/\\/|git:\\/\\/|svn:\\/\\/|sftp:"
    "\\/\\/|ftp:\\/\\/)?[a-z0-9]+([\\-\\.]{1}[a-z0-9]+){0,100}\\.[a-z]{2,5}(:[0-9]{1,5})?(\\/.*)?"
)
supported_download_repos: str = "(git|hg|svn|bzr)"
git_pattern = "(git\\+git@[a-zA-Z0-9\\.\\-]+:[a-zA-Z0-9/\\\\.@\\-]+)"
bazaar_pattern = "(bzr\\+lp:[a-zA-Z0-9\\.\\-]+)"
download_location_pattern = (
    "^(((" + supported_download_repos + "\\+)?" + url_pattern + ")|" + git_pattern + "|" + bazaar_pattern + ")$"
)


def validate_url(url: str) -> List[str]:
    if not re.match(url_pattern, url):
        return [f"must be a valid URL, but is: {url}"]

    return []


def validate_download_location(location: str) -> List[str]:
    if not re.match(download_location_pattern, location):
        return [f"must be a valid download location according to the specification, but is: {location}"]

    return []


def validate_uri(uri: str) -> List[str]:
    if not isabsuri(uri):
        return [f"must be a valid URI specified in RFC-3986 and must contain no fragment (#), but is: {uri}"]
    else:
        split = urisplit(uri)
        if split.scheme is None:
            return [f"must have a URI scheme, but is: {uri}"]

    return []



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import Dict

from spdx_tools.spdx.model import Document
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.jsonlikedict.annotation_parser import AnnotationParser
from spdx_tools.spdx.parser.jsonlikedict.creation_info_parser import CreationInfoParser
from spdx_tools.spdx.parser.jsonlikedict.dict_parsing_functions import parse_list_of_elements
from spdx_tools.spdx.parser.jsonlikedict.extracted_licensing_info_parser import ExtractedLicensingInfoParser
from spdx_tools.spdx.parser.jsonlikedict.file_parser import FileParser
from spdx_tools.spdx.parser.jsonlikedict.package_parser import PackageParser
from spdx_tools.spdx.parser.jsonlikedict.relationship_parser import RelationshipParser
from spdx_tools.spdx.parser.jsonlikedict.snippet_parser import SnippetParser
from spdx_tools.spdx.parser.logger import Logger
from spdx_tools.spdx.parser.parsing_functions import (
    construct_or_raise_parsing_error,
    raise_parsing_error_if_logger_has_messages,
)


class JsonLikeDictParser:
    logger: Logger
    creation_info_parser: CreationInfoParser
    package_parser: PackageParser
    file_parser: FileParser
    snippet_parser: SnippetParser
    extracted_licensing_info_parser: ExtractedLicensingInfoParser
    relationship_parser: RelationshipParser
    annotation_parser: AnnotationParser

    def __init__(self):
        self.logger = Logger()
        self.creation_info_parser = CreationInfoParser()
        self.package_parser = PackageParser()
        self.file_parser = FileParser()
        self.snippet_parser = SnippetParser()
        self.extracted_licensing_info_parser = ExtractedLicensingInfoParser()
        self.relationship_parser = RelationshipParser()
        self.annotation_parser = AnnotationParser()

    def parse(self, json_like_dict: Dict) -> Document:
        fields_to_parse = [
            ("creation_info", json_like_dict, self.creation_info_parser.parse_creation_info, False),
            (
                "packages",
                json_like_dict.get("packages"),
                lambda x: parse_list_of_elements(x, self.package_parser.parse_package, self.package_parser.logger),
                True,
            ),
            (
                "files",
                json_like_dict.get("files"),
                lambda x: parse_list_of_elements(x, self.file_parser.parse_file, self.file_parser.logger),
                True,
            ),
            ("annotations", json_like_dict, self.annotation_parser.parse_all_annotations, True),
            (
                "snippets",
                json_like_dict.get("snippets"),
                lambda x: parse_list_of_elements(x, self.snippet_parser.parse_snippet, self.snippet_parser.logger),
                True,
            ),
            ("relationships", json_like_dict, self.relationship_parser.parse_all_relationships, True),
            (
                "extracted_licensing_info",
                json_like_dict.get("hasExtractedLicensingInfos"),
                lambda x: parse_list_of_elements(
                    x,
                    self.extracted_licensing_info_parser.parse_extracted_licensing_info,
                    self.extracted_licensing_info_parser.logger,
                ),
                True,
            ),
        ]

        parsed_fields = {}

        for argument_name, field, parsing_method, optional in fields_to_parse:
            if optional and not field:
                continue
            try:
                parsed_fields[argument_name] = parsing_method(field)
            except SPDXParsingError as err:
                self.logger.extend(err.get_messages())

        raise_parsing_error_if_logger_has_messages(self.logger)

        document = construct_or_raise_parsing_error(Document, parsed_fields)

        return document



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from datetime import datetime

from license_expression import get_spdx_licensing

from spdx_tools.spdx.constants import DOCUMENT_SPDX_ID
from spdx_tools.spdx.model import (
    Actor,
    ActorType,
    Annotation,
    AnnotationType,
    Checksum,
    ChecksumAlgorithm,
    CreationInfo,
    Document,
    ExternalDocumentRef,
    ExternalPackageRef,
    ExternalPackageRefCategory,
    ExtractedLicensingInfo,
    File,
    FileType,
    Package,
    PackagePurpose,
    PackageVerificationCode,
    Relationship,
    RelationshipType,
    Snippet,
    SpdxNoAssertion,
    SpdxNone,
    Version,
)

# Utility methods to create data model instances. All properties have valid defaults, so they don't need to be
# specified unless relevant for the test.


def actor_fixture(actor_type=ActorType.PERSON, name="actorName", email="some@mail.com") -> Actor:
    return Actor(actor_type, name, email)


def checksum_fixture(algorithm=ChecksumAlgorithm.SHA1, value="71c4025dd9897b364f3ebbb42c484ff43d00791c") -> Checksum:
    return Checksum(algorithm, value)


def package_verification_code_fixture(
    value="85ed0817af83a24ad8da68c2b5094de69833983c", excluded_files=None
) -> PackageVerificationCode:
    excluded_files = ["./exclude.py"] if excluded_files is None else excluded_files
    return PackageVerificationCode(value, excluded_files)


def creation_info_fixture(
    spdx_version="SPDX-2.3",
    spdx_id=DOCUMENT_SPDX_ID,
    name="documentName",
    document_namespace="https://some.namespace",
    creators=None,
    created=datetime(2022, 12, 1),
    creator_comment="creatorComment",
    data_license="CC0-1.0",
    external_document_refs=None,
    license_list_version=Version(3, 19),
    document_comment="documentComment",
) -> CreationInfo:
    creators = [actor_fixture(name="creatorName")] if creators is None else creators
    external_document_refs = (
        [external_document_ref_fixture()] if external_document_refs is None else external_document_refs
    )
    return CreationInfo(
        spdx_version,
        spdx_id,
        name,
        document_namespace,
        creators,
        created,
        creator_comment,
        data_license,
        external_document_refs,
        license_list_version,
        document_comment,
    )


def file_fixture(
    name="./fileName.py",
    spdx_id="SPDXRef-File",
    checksums=None,
    file_types=None,
    license_concluded=get_spdx_licensing().parse("MIT and GPL-2.0"),
    license_info_in_file=None,
    license_comment="licenseComment",
    copyright_text="copyrightText",
    comment="fileComment",
    notice="fileNotice",
    contributors=None,
    attribution_texts=None,
) -> File:
    checksums = [checksum_fixture()] if checksums is None else checksums
    file_types = [FileType.TEXT] if file_types is None else file_types
    license_info_in_file = (
        [get_spdx_licensing().parse("MIT"), get_spdx_licensing().parse("GPL-2.0"), SpdxNoAssertion()]
        if license_info_in_file is None
        else license_info_in_file
    )
    contributors = ["fileContributor"] if contributors is None else contributors
    attribution_texts = ["fileAttributionText"] if attribution_texts is None else attribution_texts
    return File(
        name=name,
        spdx_id=spdx_id,
        checksums=checksums,
        file_types=file_types,
        license_concluded=license_concluded,
        license_info_in_file=license_info_in_file,
        license_comment=license_comment,
        copyright_text=copyright_text,
        comment=comment,
        notice=notice,
        contributors=contributors,
        attribution_texts=attribution_texts,
    )


def package_fixture(
    spdx_id="SPDXRef-Package",
    name="packageName",
    download_location="https://download.com",
    version="12.2",
    file_name="./packageFileName",
    supplier=actor_fixture(name="supplierName"),
    originator=actor_fixture(name="originatorName"),
    files_analyzed=True,
    verification_code=package_verification_code_fixture(),
    checksums=None,
    homepage="https://homepage.com",
    source_info="sourceInfo",
    license_concluded=get_spdx_licensing().parse("MIT and GPL-2.0"),
    license_info_from_files=None,
    license_declared=get_spdx_licensing().parse("MIT and GPL-2.0"),
    license_comment="packageLicenseComment",
    copyright_text="packageCopyrightText",
    summary="packageSummary",
    description="packageDescription",
    comment="packageComment",
    external_references=None,
    attribution_texts=None,
    primary_package_purpose=PackagePurpose.SOURCE,
    release_date=datetime(2022, 11, 1),
    built_date=datetime(2022, 11, 2),
    valid_until_date=datetime(2022, 11, 3),
) -> Package:
    checksums = [checksum_fixture()] if checksums is None else checksums
    license_info_from_files = (
        [get_spdx_licensing().parse("MIT"), get_spdx_licensing().parse("GPL-2.0"), SpdxNoAssertion()]
        if license_info_from_files is None
        else license_info_from_files
    )
    external_references = [external_package_ref_fixture()] if external_references is None else external_references
    attribution_texts = ["packageAttributionText"] if attribution_texts is None else attribution_texts
    return Package(
        spdx_id=spdx_id,
        name=name,
        download_location=download_location,
        version=version,
        file_name=file_name,
        supplier=supplier,
        originator=originator,
        files_analyzed=files_analyzed,
        verification_code=verification_code,
        checksums=checksums,
        homepage=homepage,
        source_info=source_info,
        license_concluded=license_concluded,
        license_info_from_files=license_info_from_files,
        license_declared=license_declared,
        license_comment=license_comment,
        copyright_text=copyright_text,
        summary=summary,
        description=description,
        comment=comment,
        external_references=external_references,
        attribution_texts=attribution_texts,
        primary_package_purpose=primary_package_purpose,
        release_date=release_date,
        built_date=built_date,
        valid_until_date=valid_until_date,
    )


def external_document_ref_fixture(
    document_ref_id="DocumentRef-external", document_uri="https://namespace.com", checksum=checksum_fixture()
) -> ExternalDocumentRef:
    return ExternalDocumentRef(document_ref_id=document_ref_id, document_uri=document_uri, checksum=checksum)


def external_package_ref_fixture(
    category=ExternalPackageRefCategory.PACKAGE_MANAGER,
    reference_type="maven-central",
    locator="org.apache.tomcat:tomcat:9.0.0.M4",
    comment="externalPackageRefComment",
) -> ExternalPackageRef:
    return ExternalPackageRef(category=category, reference_type=reference_type, locator=locator, comment=comment)


def snippet_fixture(
    spdx_id="SPDXRef-Snippet",
    file_spdx_id="SPDXRef-File",
    byte_range=(1, 2),
    line_range=(3, 4),
    license_concluded=get_spdx_licensing().parse("MIT and GPL-2.0"),
    license_info_in_snippet=None,
    license_comment="snippetLicenseComment",
    copyright_text="licenseCopyrightText",
    comment="snippetComment",
    name="snippetName",
    attribution_texts=None,
) -> Snippet:
    license_info_in_snippet = (
        [get_spdx_licensing().parse("MIT"), get_spdx_licensing().parse("GPL-2.0"), SpdxNone()]
        if license_info_in_snippet is None
        else license_info_in_snippet
    )
    attribution_texts = ["snippetAttributionText"] if attribution_texts is None else attribution_texts
    return Snippet(
        spdx_id=spdx_id,
        file_spdx_id=file_spdx_id,
        byte_range=byte_range,
        line_range=line_range,
        license_concluded=license_concluded,
        license_info_in_snippet=license_info_in_snippet,
        license_comment=license_comment,
        copyright_text=copyright_text,
        comment=comment,
        name=name,
        attribution_texts=attribution_texts,
    )


def annotation_fixture(
    spdx_id="SPDXRef-File",
    annotation_type=AnnotationType.REVIEW,
    annotator=actor_fixture(name="annotatorName"),
    annotation_date=datetime(2022, 12, 24),
    annotation_comment="annotationComment",
) -> Annotation:
    return Annotation(
        spdx_id=spdx_id,
        annotation_type=annotation_type,
        annotator=annotator,
        annotation_date=annotation_date,
        annotation_comment=annotation_comment,
    )


def extracted_licensing_info_fixture(
    license_id="LicenseRef-1",
    extracted_text="extractedText",
    license_name="licenseName",
    cross_references=None,
    comment="licenseComment",
) -> ExtractedLicensingInfo:
    cross_references = ["https://see.also"] if cross_references is None else cross_references
    return ExtractedLicensingInfo(
        license_id=license_id,
        extracted_text=extracted_text,
        license_name=license_name,
        cross_references=cross_references,
        comment=comment,
    )


def relationship_fixture(
    spdx_element_id=DOCUMENT_SPDX_ID,
    relationship_type=RelationshipType.DESCRIBES,
    related_spdx_element_id="SPDXRef-File",
    comment="relationshipComment",
) -> Relationship:
    return Relationship(
        spdx_element_id=spdx_element_id,
        relationship_type=relationship_type,
        related_spdx_element_id=related_spdx_element_id,
        comment=comment,
    )


def document_fixture(
    creation_info=None,
    packages=None,
    files=None,
    snippets=None,
    annotations=None,
    relationships=None,
    extracted_licensing_info=None,
) -> Document:
    creation_info = creation_info_fixture() if creation_info is None else creation_info
    packages = [package_fixture()] if packages is None else packages
    files = [file_fixture()] if files is None else files
    snippets = [snippet_fixture()] if snippets is None else snippets
    annotations = [annotation_fixture()] if annotations is None else annotations
    relationships = [relationship_fixture()] if relationships is None else relationships
    extracted_licensing_info = (
        [extracted_licensing_info_fixture()] if extracted_licensing_info is None else extracted_licensing_info
    )
    return Document(
        creation_info=creation_info,
        packages=packages,
        files=files,
        snippets=snippets,
        annotations=annotations,
        relationships=relationships,
        extracted_licensing_info=extracted_licensing_info,
    )



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import xmltodict
from beartype.typing import IO

from spdx_tools.spdx.jsonschema.document_converter import DocumentConverter
from spdx_tools.spdx.model import Document
from spdx_tools.spdx.writer.write_utils import convert, validate_and_deduplicate


def write_document_to_stream(
    document: Document,
    stream: IO[str],
    validate: bool = True,
    converter: DocumentConverter = None,
    drop_duplicates: bool = True,
):
    """
    Serializes the provided document to XML and writes it to a file with the provided name. Unless validate is set
    to False, validates the document before serialization. Unless a DocumentConverter instance is provided,
    a new one is created.
    """
    document = validate_and_deduplicate(document, validate, drop_duplicates)
    document_dict = {"Document": convert(document, converter)}
    xmltodict.unparse(document_dict, stream, encoding="utf-8", pretty=True)


def write_document_to_file(
    document: Document,
    file_name: str,
    validate: bool = True,
    converter: DocumentConverter = None,
    drop_duplicates: bool = True,
):
    with open(file_name, "w") as out:
        write_document_to_stream(document, out, validate, converter, drop_duplicates)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from unittest.mock import MagicMock, call, mock_open, patch

import pytest

from spdx_tools.spdx.model import SpdxNoAssertion, SpdxNone
from spdx_tools.spdx.writer.tagvalue.relationship_writer import write_relationship
from tests.spdx.fixtures import relationship_fixture


@pytest.mark.parametrize(
    "relationship, expected_calls",
    [
        (
            relationship_fixture(),
            [
                call("Relationship: SPDXRef-DOCUMENT DESCRIBES SPDXRef-File\n"),
                call("RelationshipComment: relationshipComment\n"),
            ],
        ),
        (
            relationship_fixture(related_spdx_element_id=SpdxNoAssertion(), comment=None),
            [call("Relationship: SPDXRef-DOCUMENT DESCRIBES NOASSERTION\n")],
        ),
        (
            relationship_fixture(
                spdx_element_id="DocumentRef-External:SPDXRef-DOCUMENT",
                related_spdx_element_id=SpdxNone(),
                comment=None,
            ),
            [call("Relationship: DocumentRef-External:SPDXRef-DOCUMENT DESCRIBES NONE\n")],
        ),
    ],
)
def test_relationship_writer(relationship, expected_calls):
    mock: MagicMock = mock_open()
    with patch(f"{__name__}.open", mock, create=True):
        with open("foo", "w") as file:
            write_relationship(relationship, file)

    mock.assert_called_once_with("foo", "w")
    handle = mock()
    handle.write.assert_has_calls(expected_calls)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import TextIO

from spdx_tools.spdx3.model import Hash
from spdx_tools.spdx3.writer.console.console import write_value
from spdx_tools.spdx3.writer.console.integrity_method_writer import write_integrity_method


def write_hash(hash_object: Hash, text_output: TextIO, heading: bool, indent: bool = True):
    if heading:
        text_output.write("## Hash\n")
    write_value("algorithm", hash_object.algorithm, text_output, indent)
    write_value("hash_value", hash_object.hash_value, text_output, indent)
    write_integrity_method(hash_object, text_output, indent)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import TextIO

from spdx_tools.spdx3.model import Bom
from spdx_tools.spdx3.writer.console.bundle_writer import write_bundle


def write_bom(bom: Bom, text_output: TextIO, heading: bool = True):
    if heading:
        text_output.write("## Bom\n")
    write_bundle(bom, text_output, False)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from datetime import datetime
from enum import Enum

from beartype.typing import Any, List
from semantic_version import Version

from spdx_tools.spdx3.model.creation_info import CreationInfo
from spdx_tools.spdx3.model.hash import Hash
from spdx_tools.spdx3.payload import Payload
from spdx_tools.spdx.casing_tools import snake_case_to_camel_case
from spdx_tools.spdx.datetime_conversions import datetime_to_iso_string


def convert_payload_to_json_ld_list_of_elements(payload: Payload) -> List:
    element_list = []

    for element in payload.get_full_map().values():
        element_dict = _convert_to_json_ld_dict(element)
        element_list.append(element_dict)

    return element_list


def _convert_to_json_ld_dict(element: Any, alt_creation_info=False, alt_hash=False):
    if not element:
        return None

    if isinstance(element, (str, int, tuple)):
        return element

    if isinstance(element, Version):
        return str(element)

    if isinstance(element, datetime):
        return datetime_to_iso_string(element)

    if isinstance(element, Enum):
        return snake_case_to_camel_case(element.name)

    if isinstance(element, list):
        return [_convert_to_json_ld_dict(item) for item in element if item]

    if alt_hash and isinstance(element, Hash):
        hash_dict = {element.algorithm.name: element.hash_value}
        if element.comment:
            hash_dict["comment"] = element.comment
        return hash_dict

    # if issubclass(element.__class__, Element):
    #     element_dict = {"@type": element.__class__.__name__}
    # else:
    #     element_dict = {}  # typing of non-Element classes should be handled by the @context, I think

    element_dict = {"@type": element.__class__.__name__}

    for attribute_name in vars(element):
        attribute_value = getattr(element, attribute_name)

        if alt_creation_info and isinstance(attribute_value, CreationInfo):
            for creation_info_attr_name in vars(attribute_value):
                creation_info_attr_value = getattr(attribute_value, creation_info_attr_name)
                element_dict[snake_case_to_camel_case(creation_info_attr_name)] = _convert_to_json_ld_dict(
                    creation_info_attr_value
                )

        elif attribute_value:
            if attribute_name == "_spdx_id":
                attribute_name = "@id"
            elif attribute_name == "_from_element":
                attribute_name = "from"
            else:
                attribute_name = snake_case_to_camel_case(attribute_name)

            element_dict[attribute_name] = _convert_to_json_ld_dict(attribute_value)

    return element_dict



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from datetime import datetime

from beartype.typing import Dict, List, Optional, Union
from license_expression import LicenseExpression

from spdx_tools.spdx.datetime_conversions import datetime_from_str
from spdx_tools.spdx.model import (
    Actor,
    ExternalPackageRef,
    ExternalPackageRefCategory,
    Package,
    PackagePurpose,
    PackageVerificationCode,
    SpdxNoAssertion,
    SpdxNone,
)
from spdx_tools.spdx.parser.actor_parser import ActorParser
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.jsonlikedict.checksum_parser import ChecksumParser
from spdx_tools.spdx.parser.jsonlikedict.dict_parsing_functions import (
    append_parsed_field_or_log_error,
    json_str_to_enum_name,
    parse_field_or_log_error,
    parse_field_or_no_assertion,
    parse_field_or_no_assertion_or_none,
)
from spdx_tools.spdx.parser.jsonlikedict.license_expression_parser import LicenseExpressionParser
from spdx_tools.spdx.parser.logger import Logger
from spdx_tools.spdx.parser.parsing_functions import (
    construct_or_raise_parsing_error,
    raise_parsing_error_if_logger_has_messages,
)


class PackageParser:
    logger: Logger
    actor_parser: ActorParser
    checksum_parser: ChecksumParser
    license_expression_parser: LicenseExpressionParser

    def __init__(self):
        self.actor_parser = ActorParser()
        self.checksum_parser = ChecksumParser()
        self.license_expression_parser = LicenseExpressionParser()
        self.logger = Logger()

    def parse_package(self, package_dict: Dict) -> Package:
        logger = Logger()
        name: Optional[str] = package_dict.get("name")
        spdx_id: Optional[str] = package_dict.get("SPDXID")
        attribution_texts: List[str] = package_dict.get("attributionTexts", [])

        built_date: Optional[datetime] = parse_field_or_log_error(
            logger, package_dict.get("builtDate"), datetime_from_str
        )

        checksums = parse_field_or_log_error(
            logger, package_dict.get("checksums"), self.checksum_parser.parse_checksum, field_is_list=True
        )
        comment: Optional[str] = package_dict.get("comment")
        copyright_text: Optional[Union[str, SpdxNoAssertion, SpdxNone]] = parse_field_or_no_assertion_or_none(
            package_dict.get("copyrightText")
        )
        description: Optional[str] = package_dict.get("description")
        download_location: Optional[Union[str, SpdxNoAssertion, SpdxNone]] = parse_field_or_no_assertion_or_none(
            package_dict.get("downloadLocation")
        )

        external_refs: List[ExternalPackageRef] = parse_field_or_log_error(
            logger, package_dict.get("externalRefs"), self.parse_external_refs
        )

        files_analyzed: Optional[Union[bool, str]] = package_dict.get("filesAnalyzed")

        if files_analyzed is None:  # default value is True
            files_analyzed = True
        elif isinstance(files_analyzed, str):  # XML does not support boolean typed values
            if files_analyzed.lower() == "true":
                files_analyzed = True
            elif files_analyzed.lower() == "false":
                files_analyzed = False

        homepage: Optional[Union[str, SpdxNoAssertion, SpdxNone]] = parse_field_or_no_assertion_or_none(
            package_dict.get("homepage")
        )
        license_comments: Optional[str] = package_dict.get("licenseComments")
        license_concluded = parse_field_or_log_error(
            logger, package_dict.get("licenseConcluded"), self.license_expression_parser.parse_license_expression
        )

        license_declared: Optional[Union[LicenseExpression, SpdxNoAssertion, SpdxNone]] = parse_field_or_log_error(
            logger, package_dict.get("licenseDeclared"), self.license_expression_parser.parse_license_expression
        )

        license_info_from_file: List[Union[LicenseExpression, SpdxNoAssertion, SpdxNone]] = parse_field_or_log_error(
            logger,
            package_dict.get("licenseInfoFromFiles"),
            self.license_expression_parser.parse_license_expression,
            field_is_list=True,
        )
        originator: Optional[Union[Actor, SpdxNoAssertion]] = parse_field_or_log_error(
            logger,
            package_dict.get("originator"),
            lambda x: parse_field_or_no_assertion(x, self.actor_parser.parse_actor),
        )
        package_file_name: Optional[str] = package_dict.get("packageFileName")

        package_verification_code: Optional[PackageVerificationCode] = parse_field_or_log_error(
            logger, package_dict.get("packageVerificationCode"), self.parse_package_verification_code
        )
        primary_package_purpose: Optional[PackagePurpose] = parse_field_or_log_error(
            logger, package_dict.get("primaryPackagePurpose"), self.parse_primary_package_purpose
        )

        release_date: Optional[datetime] = parse_field_or_log_error(
            logger, package_dict.get("releaseDate"), datetime_from_str
        )
        source_info: Optional[str] = package_dict.get("sourceInfo")
        summary: Optional[str] = package_dict.get("summary")
        supplier: Optional[Union[Actor, SpdxNoAssertion]] = parse_field_or_log_error(
            logger,
            package_dict.get("supplier"),
            lambda x: parse_field_or_no_assertion(x, self.actor_parser.parse_actor),
        )
        valid_until_date: Optional[datetime] = parse_field_or_log_error(
            logger, package_dict.get("validUntilDate"), datetime_from_str
        )

        version_info: Optional[str] = package_dict.get("versionInfo")
        raise_parsing_error_if_logger_has_messages(logger, "Package")

        package = construct_or_raise_parsing_error(
            Package,
            dict(
                spdx_id=spdx_id,
                name=name,
                download_location=download_location,
                version=version_info,
                file_name=package_file_name,
                supplier=supplier,
                originator=originator,
                files_analyzed=files_analyzed,
                verification_code=package_verification_code,
                checksums=checksums,
                homepage=homepage,
                source_info=source_info,
                license_concluded=license_concluded,
                license_info_from_files=license_info_from_file,
                license_declared=license_declared,
                license_comment=license_comments,
                copyright_text=copyright_text,
                summary=summary,
                description=description,
                comment=comment,
                external_references=external_refs,
                attribution_texts=attribution_texts,
                primary_package_purpose=primary_package_purpose,
                release_date=release_date,
                built_date=built_date,
                valid_until_date=valid_until_date,
            ),
        )

        return package

    def parse_external_refs(self, external_ref_dicts: List[Dict]) -> List[ExternalPackageRef]:
        external_refs = []
        for external_ref_dict in external_ref_dicts:
            external_refs = append_parsed_field_or_log_error(
                self.logger, external_refs, external_ref_dict, self.parse_external_ref
            )
        return external_refs

    def parse_external_ref(self, external_ref_dict: Dict) -> ExternalPackageRef:
        logger = Logger()
        ref_category = parse_field_or_log_error(
            logger, external_ref_dict.get("referenceCategory"), self.parse_external_ref_category
        )
        ref_locator: Optional[str] = external_ref_dict.get("referenceLocator")
        ref_type: Optional[str] = external_ref_dict.get("referenceType")
        comment: Optional[str] = external_ref_dict.get("comment")
        raise_parsing_error_if_logger_has_messages(logger, "ExternalPackageRef")
        external_ref = construct_or_raise_parsing_error(
            ExternalPackageRef,
            dict(category=ref_category, reference_type=ref_type, locator=ref_locator, comment=comment),
        )

        return external_ref

    @staticmethod
    def parse_external_ref_category(external_ref_category_str: str) -> ExternalPackageRefCategory:
        try:
            external_ref_category = ExternalPackageRefCategory[json_str_to_enum_name(external_ref_category_str)]
        except KeyError:
            raise SPDXParsingError([f"Invalid ExternalPackageRefCategory: {external_ref_category_str}"])

        return external_ref_category

    @staticmethod
    def parse_package_verification_code(verification_code_dict: Dict) -> PackageVerificationCode:
        excluded_files: List[str] = verification_code_dict.get("packageVerificationCodeExcludedFiles", [])
        verification_code_value: Optional[str] = verification_code_dict.get("packageVerificationCodeValue")

        package_verification_code = construct_or_raise_parsing_error(
            PackageVerificationCode, dict(value=verification_code_value, excluded_files=excluded_files)
        )

        return package_verification_code

    @staticmethod
    def parse_primary_package_purpose(primary_package_purpose: str) -> PackagePurpose:
        try:
            return PackagePurpose[json_str_to_enum_name(primary_package_purpose)]
        except KeyError:
            raise SPDXParsingError([f"Invalid PrimaryPackagePurpose: {primary_package_purpose}"])



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from unittest import TestCase

import pytest

from spdx_tools.spdx.document_utils import (
    create_document_without_duplicates,
    create_list_without_duplicates,
    get_contained_spdx_element_ids,
    get_contained_spdx_elements,
    get_element_from_spdx_id,
)
from spdx_tools.spdx.model import FileType, SpdxNoAssertion, SpdxNone
from tests.spdx.fixtures import (
    actor_fixture,
    checksum_fixture,
    creation_info_fixture,
    document_fixture,
    external_document_ref_fixture,
    external_package_ref_fixture,
    extracted_licensing_info_fixture,
    file_fixture,
    package_fixture,
    snippet_fixture,
)


@pytest.fixture
def variables():
    return document_fixture(), package_fixture(), file_fixture(), snippet_fixture()


def test_contained_element_ids(variables):
    document, package, file, snippet = variables
    element_ids = get_contained_spdx_element_ids(document)
    TestCase().assertCountEqual(element_ids, [package.spdx_id, file.spdx_id, snippet.spdx_id])


def test_get_element_from_spdx_id(variables):
    document, package, file, snippet = variables
    assert get_element_from_spdx_id(document, package.spdx_id) == package
    assert get_element_from_spdx_id(document, file.spdx_id) == file
    assert get_element_from_spdx_id(document, snippet.spdx_id) == snippet
    assert get_element_from_spdx_id(document, "unknown_id") is None


def test_get_contained_spdx_elements(variables):
    document, package, file, snippet = variables
    contained_elements = get_contained_spdx_elements(document)
    assert contained_elements[package.spdx_id] == package
    assert contained_elements[file.spdx_id] == file
    assert contained_elements[snippet.spdx_id] == snippet


def test_create_list_without_duplicates():
    list_with_duplicates = [1, 2, 3, 5, 1, 67, 9, 67]

    list_without_duplicates = create_list_without_duplicates(list_with_duplicates)

    assert list_without_duplicates == [1, 2, 3, 5, 67, 9]


def test_create_document_without_duplicates():
    document = document_fixture(
        creation_info=creation_info_fixture(
            creators=[actor_fixture(name="creatorName"), actor_fixture(name="creatorName")],
            external_document_refs=[external_document_ref_fixture(), external_document_ref_fixture()],
        ),
        packages=[
            package_fixture(
                checksums=[checksum_fixture(), checksum_fixture()],
                license_info_from_files=[SpdxNoAssertion(), SpdxNoAssertion()],
                external_references=[external_package_ref_fixture(), external_package_ref_fixture()],
                attribution_texts=["duplicated text", "duplicated text"],
            )
        ],
        files=[
            file_fixture(
                checksums=[checksum_fixture(), checksum_fixture()],
                file_types=[FileType.TEXT, FileType.TEXT],
                license_info_in_file=[SpdxNoAssertion(), SpdxNoAssertion()],
                contributors=["duplicated contributor", "duplicated contributor"],
                attribution_texts=["duplicated text", "duplicated text"],
            )
        ],
        snippets=[
            snippet_fixture(
                license_info_in_snippet=[SpdxNone(), SpdxNone()],
                attribution_texts=["duplicated text", "duplicated text"],
            )
        ],
        extracted_licensing_info=[
            extracted_licensing_info_fixture(cross_references=["duplicated reference", "duplicated reference"])
        ],
    )
    expected_document = document_fixture(
        creation_info=creation_info_fixture(
            creators=[actor_fixture(name="creatorName")], external_document_refs=[external_document_ref_fixture()]
        ),
        packages=[
            package_fixture(
                checksums=[checksum_fixture()],
                license_info_from_files=[SpdxNoAssertion()],
                external_references=[external_package_ref_fixture()],
                attribution_texts=["duplicated text"],
            )
        ],
        files=[
            file_fixture(
                checksums=[checksum_fixture()],
                file_types=[FileType.TEXT],
                license_info_in_file=[SpdxNoAssertion()],
                contributors=["duplicated contributor"],
                attribution_texts=["duplicated text"],
            )
        ],
        snippets=[snippet_fixture(license_info_in_snippet=[SpdxNone()], attribution_texts=["duplicated text"])],
        extracted_licensing_info=[extracted_licensing_info_fixture(cross_references=["duplicated reference"])],
    )

    document_without_duplicates = create_document_without_duplicates(document)

    assert document_without_duplicates == expected_document



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import List, Optional

from spdx_tools.common.typing.dataclass_with_properties import dataclass_with_properties
from spdx_tools.common.typing.type_checks import check_types_and_set_values
from spdx_tools.spdx3.model.licensing.license import License


@dataclass_with_properties
class CustomLicense(License):
    def __init__(
        self,
        license_id: str,
        license_name: str,
        license_text: str,
        license_comment: Optional[str] = None,
        see_also: List[str] = None,
        is_osi_approved: Optional[bool] = None,
        is_fsf_libre: Optional[bool] = None,
        standard_license_header: Optional[str] = None,
        standard_license_template: Optional[str] = None,
        is_deprecated_license_id: Optional[bool] = None,
        obsoleted_by: Optional[str] = None,
    ):
        see_also = [] if see_also is None else see_also
        check_types_and_set_values(self, locals())



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import Union
from license_expression import ExpressionError, LicenseExpression, Licensing

from spdx_tools.spdx.model import SpdxNoAssertion, SpdxNone
from spdx_tools.spdx.parser.error import SPDXParsingError


class LicenseExpressionParser:
    @staticmethod
    def parse_license_expression(license_expression_str: str) -> Union[LicenseExpression, SpdxNone, SpdxNoAssertion]:
        if isinstance(license_expression_str, str):
            if license_expression_str.upper() == "NOASSERTION":
                return SpdxNoAssertion()
            if license_expression_str.upper() == "NONE":
                return SpdxNone()

        try:
            license_expression = Licensing().parse(license_expression_str)
        except ExpressionError as err:
            raise SPDXParsingError([f"Error parsing LicenseExpression: {err.args[0]}: {license_expression_str}"])

        return license_expression



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from abc import abstractmethod

from beartype.typing import Optional

from spdx_tools.common.typing.dataclass_with_properties import dataclass_with_properties
from spdx_tools.spdx3.model.security.vuln_assessment_relationship import VulnAssessmentRelationship


@dataclass_with_properties
class VexVulnAssessmentRelationship(VulnAssessmentRelationship):
    vex_version: Optional[str] = None
    status_notes: Optional[str] = None

    @abstractmethod
    def __init__(self):
        pass



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import pytest

from spdx_tools.spdx3.bump_from_spdx2.checksum import bump_checksum, convert_checksum_algorithm_to_hash_algorithm
from spdx_tools.spdx3.model import HashAlgorithm
from spdx_tools.spdx.model.checksum import ChecksumAlgorithm
from tests.spdx.fixtures import checksum_fixture


def test_bump_checksum():
    checksum = checksum_fixture()
    hash = bump_checksum(checksum)

    assert hash.algorithm == HashAlgorithm.SHA1
    assert hash.hash_value == "71c4025dd9897b364f3ebbb42c484ff43d00791c"


@pytest.mark.parametrize(
    "checksum_algorithm,expected_hash_algorithm",
    [
        (ChecksumAlgorithm.SHA1, HashAlgorithm.SHA1),
        (ChecksumAlgorithm.SHA224, HashAlgorithm.SHA224),
        (ChecksumAlgorithm.SHA256, HashAlgorithm.SHA256),
        (ChecksumAlgorithm.SHA384, HashAlgorithm.SHA384),
        (ChecksumAlgorithm.SHA512, HashAlgorithm.SHA512),
        (ChecksumAlgorithm.SHA3_256, HashAlgorithm.SHA3_256),
        (ChecksumAlgorithm.SHA3_384, HashAlgorithm.SHA3_384),
        (ChecksumAlgorithm.SHA3_512, HashAlgorithm.SHA3_512),
        (ChecksumAlgorithm.BLAKE2B_256, HashAlgorithm.BLAKE2B256),
        (ChecksumAlgorithm.BLAKE2B_384, HashAlgorithm.BLAKE2B384),
        (ChecksumAlgorithm.BLAKE2B_512, HashAlgorithm.BLAKE2B512),
        (ChecksumAlgorithm.BLAKE3, HashAlgorithm.BLAKE3),
        (ChecksumAlgorithm.MD2, HashAlgorithm.MD2),
        (ChecksumAlgorithm.MD4, HashAlgorithm.MD4),
        (ChecksumAlgorithm.MD5, HashAlgorithm.MD5),
        (ChecksumAlgorithm.MD6, HashAlgorithm.MD6),
        (ChecksumAlgorithm.ADLER32, HashAlgorithm.OTHER),
    ],
)
def test_bump_checksum_algorithm(checksum_algorithm, expected_hash_algorithm):
    hash_algorithm = convert_checksum_algorithm_to_hash_algorithm(checksum_algorithm)

    assert hash_algorithm == expected_hash_algorithm



# SPDX-License-Identifier: Apache-2.0
#  Copyright (c) 2022 spdx contributors
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#    http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from beartype.typing import TextIO

from spdx_tools.spdx.model import Snippet
from spdx_tools.spdx.writer.tagvalue.tagvalue_writer_helper_functions import write_range, write_text_value, write_value


def write_snippet(snippet: Snippet, text_output: TextIO):
    text_output.write("## Snippet Information\n")

    write_value("SnippetSPDXID", snippet.spdx_id, text_output)
    write_value("SnippetFromFileSPDXID", snippet.file_spdx_id, text_output)
    write_range("SnippetByteRange", snippet.byte_range, text_output)
    write_range("SnippetLineRange", snippet.line_range, text_output)

    write_value("SnippetLicenseConcluded", snippet.license_concluded, text_output)
    for license_info in snippet.license_info_in_snippet:
        write_value("LicenseInfoInSnippet", license_info, text_output)
    write_text_value("SnippetLicenseComments", snippet.license_comment, text_output)
    write_text_value("SnippetCopyrightText", snippet.copyright_text, text_output)

    write_text_value("SnippetComment", snippet.comment, text_output)
    write_value("SnippetName", snippet.name, text_output)

    for attribution_text in snippet.attribution_texts:
        write_text_value("SnippetAttributionText", attribution_text, text_output)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import os

from spdx_tools.spdx3.bump_from_spdx2.spdx_document import bump_spdx_document
from spdx_tools.spdx3.payload import Payload
from spdx_tools.spdx3.writer.json_ld.json_ld_writer import write_payload
from spdx_tools.spdx.model.document import Document as Spdx2_Document
from tests.spdx.fixtures import document_fixture


def test_json_writer():
    spdx2_document: Spdx2_Document = document_fixture()
    payload: Payload = bump_spdx_document(spdx2_document)

    # this currently generates an actual file to look at, this should be changed to a temp file later
    write_payload(payload, os.path.join(os.path.dirname(__file__), "../../../SPDX3_jsonld_test"))



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from dataclasses import dataclass

from beartype import beartype
from beartype.roar import BeartypeCallHintException


def dataclass_with_properties(cls):
    """Decorator to generate a dataclass with properties out of the class' value:type list.
    Their getters and setters will be subjected to the @typechecked decorator to ensure type conformity."""
    data_cls = dataclass(cls)
    for field_name, field_type in data_cls.__annotations__.items():
        set_field = make_setter(field_name, field_type)
        get_field = make_getter(field_name, field_type)

        setattr(data_cls, field_name, property(get_field, set_field))

    return data_cls


def make_setter(field_name, field_type):
    """helper method to avoid late binding when generating functions in a for loop"""

    @beartype
    def set_field(self, value: field_type):
        setattr(self, f"_{field_name}", value)

    def set_field_with_error_conversion(self, value: field_type):
        try:
            set_field(self, value)
        except BeartypeCallHintException as err:
            error_message: str = f"SetterError {self.__class__.__name__}: {err}"

            # As setters are created dynamically, their argument name is always "value". We replace it by the
            # actual name so the error message is more helpful.
            raise TypeError(error_message.replace("value", field_name, 1) + f": {value}")

    return set_field_with_error_conversion


def make_getter(field_name, field_type):
    """helper method to avoid late binding when generating functions in a for loop"""

    def get_field(self) -> field_type:
        return getattr(self, f"_{field_name}")

    return get_field



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

from unittest import mock

import pytest

from spdx_tools.spdx.model import Checksum, ChecksumAlgorithm, File, FileType, SpdxNoAssertion, SpdxNone


@mock.patch("spdx_tools.spdx.model.Checksum", autospec=True)
def test_correct_initialization(checksum):
    file = File(
        "name",
        "id",
        [checksum, checksum],
        [FileType.OTHER, FileType.SPDX],
        SpdxNone(),
        [SpdxNoAssertion()],
        "comment on license",
        "copyright",
        "comment",
        "notice",
        ["contributor"],
        ["attribution"],
    )
    assert file.name == "name"
    assert file.spdx_id == "id"
    assert file.checksums == [checksum, checksum]
    assert file.file_types == [FileType.OTHER, FileType.SPDX]
    assert file.license_concluded == SpdxNone()
    assert file.license_info_in_file == [SpdxNoAssertion()]
    assert file.license_comment == "comment on license"
    assert file.copyright_text == "copyright"
    assert file.comment == "comment"
    assert file.notice == "notice"
    assert file.contributors == ["contributor"]
    assert file.attribution_texts == ["attribution"]


@mock.patch("spdx_tools.spdx.model.Checksum", autospec=True)
def test_correct_initialization_with_default_values(checksum):
    file = File("name", "id", [checksum, checksum])
    assert file.name == "name"
    assert file.spdx_id == "id"
    assert file.checksums == [checksum, checksum]
    assert file.file_types == []
    assert file.license_concluded is None
    assert file.license_info_in_file == []
    assert file.license_comment is None
    assert file.copyright_text is None
    assert file.comment is None
    assert file.notice is None
    assert file.contributors == []
    assert file.attribution_texts == []


@mock.patch("spdx_tools.spdx.model.Checksum", autospec=True)
def test_wrong_type_in_name(checksum):
    with pytest.raises(TypeError):
        File(42, "id", [checksum])


@mock.patch("spdx_tools.spdx.model.Checksum", autospec=True)
def test_wrong_type_in_spdx_id(checksum):
    with pytest.raises(TypeError):
        File("name", 42, [checksum])


def test_wrong_type_in_checksum():
    checksum = Checksum(ChecksumAlgorithm.BLAKE2B_256, "value")
    with pytest.raises(TypeError):
        File("name", "id", checksum)


@mock.patch("spdx_tools.spdx.model.Checksum", autospec=True)
def test_wrong_type_in_file_type(checksum):
    with pytest.raises(TypeError):
        File("name", "id", [checksum], file_types=FileType.OTHER)


@mock.patch("spdx_tools.spdx.model.Checksum", autospec=True)
def test_wrong_type_in_license_concluded(checksum):
    with pytest.raises(TypeError):
        File("name", "id", [checksum], license_concluded="NONE")


@mock.patch("spdx_tools.spdx.model.Checksum", autospec=True)
def test_wrong_type_in_license_info_in_file(checksum):
    with pytest.raises(TypeError):
        File("name", "id", [checksum], license_info_in_file=[SpdxNone])


@mock.patch("spdx_tools.spdx.model.Checksum", autospec=True)
def test_wrong_type_in_license_comment(checksum):
    with pytest.raises(TypeError):
        File("name", "id", [checksum], license_comment=42)


@mock.patch("spdx_tools.spdx.model.Checksum", autospec=True)
def test_wrong_type_in_copyright_text(checksum):
    with pytest.raises(TypeError):
        File("name", "id", [checksum], copyright_text=[SpdxNone()])


@mock.patch("spdx_tools.spdx.model.Checksum", autospec=True)
def test_wrong_type_in_comment(checksum):
    with pytest.raises(TypeError):
        File("name", "id", [checksum], comment=42)


@mock.patch("spdx_tools.spdx.model.Checksum", autospec=True)
def test_wrong_type_in_notice(checksum):
    with pytest.raises(TypeError):
        File("name", "id", [checksum], notice=["notice"])


@mock.patch("spdx_tools.spdx.model.Checksum", autospec=True)
def test_wrong_type_in_contributors(checksum):
    with pytest.raises(TypeError):
        File("name", "id", [checksum], contributors="contributor")


@mock.patch("spdx_tools.spdx.model.Checksum", autospec=True)
def test_wrong_type_in_attribution_texts(checksum):
    with pytest.raises(TypeError):
        File("name", "id", [checksum], attribution_texts=[41, 42])



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import json
import os

from spdx_tools.spdx3.payload import Payload
from spdx_tools.spdx3.writer.json_ld.json_ld_converter import convert_payload_to_json_ld_list_of_elements


def write_payload(payload: Payload, file_name: str):
    element_list = convert_payload_to_json_ld_list_of_elements(payload)

    # this will be obsolete as soon as the context is publicly available under some URI
    with open(os.path.join(os.path.dirname(__file__), "context.json"), "r") as infile:
        context = json.load(infile)

    complete_dict = {"@context": context, "@graph": element_list}

    with open(file_name + ".jsonld", "w") as out:
        json.dump(complete_dict, out, indent=2)



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import pytest

from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.jsonlikedict.extracted_licensing_info_parser import ExtractedLicensingInfoParser


def test_parse_extracted_licensing_info():
    extracted_licensing_info_parser = ExtractedLicensingInfoParser()

    extracted_licensing_infos_dict = {
        "licenseId": "LicenseRef-Beerware-4.2",
        "comment": "The beerware license has a couple of other standard variants.",
        "extractedText": '"THE BEER-WARE LICENSE" (Revision 42):\nphk@FreeBSD.ORG wrote this file. As long as you '
        "retain this notice you\ncan do whatever you want with this stuff. If we meet some day, and "
        "you think this stuff is worth it, you can buy me a beer in return Poul-Henning Kamp",
        "name": "Beer-Ware License (Version 42)",
        "seeAlsos": ["http://people.freebsd.org/~phk/"],
    }

    extracted_licensing_info = extracted_licensing_info_parser.parse_extracted_licensing_info(
        extracted_licensing_infos_dict
    )

    assert extracted_licensing_info.license_id == "LicenseRef-Beerware-4.2"
    assert extracted_licensing_info.comment == "The beerware license has a couple of other standard variants."
    assert (
        extracted_licensing_info.extracted_text
        == '"THE BEER-WARE LICENSE" (Revision 42):\nphk@FreeBSD.ORG wrote this file. As long as you retain this '
        "notice you\ncan do whatever you want with this stuff. If we meet some day, and you think this stuff is "
        "worth it, you can buy me a beer in return Poul-Henning Kamp"
    )
    assert extracted_licensing_info.license_name == "Beer-Ware License (Version 42)"
    assert extracted_licensing_info.cross_references == ["http://people.freebsd.org/~phk/"]


def test_parse_invalid_extracted_licensing_info():
    extracted_licensing_info_parser = ExtractedLicensingInfoParser()

    extracted_licensing_infos_dict = {
        "licenseId": "LicenseRef-Beerware-4.2",
        "comment": 56,
        "extractedText": '"THE BEER-WARE LICENSE" (Revision 42):\nphk@FreeBSD.ORG wrote this file. As long as you '
        "retain this notice you\ncan do whatever you want with this stuff. If we meet some day, and "
        "you think this stuff is worth it, you can buy me a beer in return Poul-Henning Kamp",
        "name": "Beer-Ware License (Version 42)",
        "seeAlsos": ["http://people.freebsd.org/~phk/"],
    }

    with pytest.raises(SPDXParsingError):
        extracted_licensing_info_parser.parse_extracted_licensing_info(extracted_licensing_infos_dict)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import pytest

from spdx_tools.spdx3.model import ExternalReference, ExternalReferenceType
from tests.spdx3.fixtures import external_reference_fixture
from tests.spdx3.model.model_test_utils import get_property_names


def test_correct_initialization():
    external_reference = external_reference_fixture()

    for property_name in get_property_names(ExternalReference):
        assert getattr(external_reference, property_name) is not None

    assert external_reference.external_reference_type == ExternalReferenceType.OTHER
    assert external_reference.locator == ["org.apache.tomcat:tomcat:9.0.0.M4"]
    assert external_reference.content_type == "externalReferenceContentType"
    assert external_reference.comment == "externalReferenceComment"


def test_invalid_initialization():
    with pytest.raises(TypeError) as err:
        ExternalReference("OTHER", "a URI", 34, True)

    assert len(err.value.args[0]) == 4
    for error in err.value.args[0]:
        assert error.startswith("SetterError ExternalReference:")



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

import pytest

from spdx_tools.spdx.model import Version


@pytest.mark.parametrize("input_string,expected", [("1.2", Version(1, 2)), ("12.345", Version(12, 345))])
def test_version_from_string(input_string, expected):
    assert Version.is_valid_version_string(input_string)
    version: Version = Version.from_string(input_string)
    assert version == expected


@pytest.mark.parametrize("input_string", ["1", "1-2", "1.a", "a", "a.b", "a.1", "v1.2", "1.2v"])
def test_invalid_version_string(input_string):
    assert not Version.is_valid_version_string(input_string)
    with pytest.raises(ValueError) as error:
        Version.from_string(input_string)
    assert str(error.value) == f"{input_string} is not a valid version string"



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from unittest import TestCase

import pytest
from license_expression import get_spdx_licensing

from spdx_tools.spdx.model import SpdxNoAssertion
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.tagvalue.parser import Parser
from tests.spdx.parser.tagvalue.test_creation_info_parser import DOCUMENT_STR


def test_parse_snippet():
    parser = Parser()
    snippet_str = "\n".join(
        [
            "SnippetSPDXID: SPDXRef-Snippet",
            "SnippetLicenseComments: <text>Some lic comment.</text>",
            "SnippetCopyrightText: <text> Copyright 2008-2010 John Smith </text>",
            "SnippetComment: <text>Some snippet comment.</text>",
            "SnippetName: from linux kernel",
            "SnippetFromFileSPDXID: SPDXRef-DoapSource",
            "SnippetLicenseConcluded: Apache-2.0",
            "LicenseInfoInSnippet: NOASSERTION",
            "SnippetByteRange: 310:420",
            "SnippetLineRange: 5:23",
            "SnippetAttributionText: <text>This is a text\nthat spans multiple lines.</text>",
            "SnippetAttributionText:   This text spans one line but has trailing and leading whitespaces.      ",
        ]
    )

    document = parser.parse("\n".join([DOCUMENT_STR, snippet_str]))
    assert document is not None
    assert len(document.snippets) == 1
    snippet = document.snippets[0]
    assert snippet.spdx_id == "SPDXRef-Snippet"
    assert snippet.name == "from linux kernel"
    assert snippet.comment == "Some snippet comment."
    assert snippet.copyright_text == " Copyright 2008-2010 John Smith "
    assert snippet.license_comment == "Some lic comment."
    assert snippet.file_spdx_id == "SPDXRef-DoapSource"
    assert snippet.license_concluded == get_spdx_licensing().parse("Apache-2.0")
    assert snippet.license_info_in_snippet == [SpdxNoAssertion()]
    assert snippet.byte_range[0] == 310
    assert snippet.byte_range[1] == 420
    assert snippet.line_range[0] == 5
    assert snippet.line_range[1] == 23
    TestCase().assertCountEqual(
        snippet.attribution_texts,
        [
            "This is a text\nthat spans multiple lines.",
            "This text spans one line but has trailing and leading whitespaces.",
        ],
    )


@pytest.mark.parametrize(
    "snippet_str, expected_message",
    [
        (
            "SnippetName: TestSnippet",
            "Element Snippet is not the current element in scope, probably the expected "
            "tag to start the element (SnippetSPDXID) is missing. Line: 1",
        ),
        (
            "SnippetSPDXID: SPDXDRef-Snippet\nSnippetByteRange: 1,4",
            "Error while parsing Snippet: [\"Value for SnippetByteRange doesn't match "
            'valid range pattern. Line: 2"]',
        ),
        (
            "SnippetSPDXID: SPDXDRef-Snippet\nSnippetByteRange: 1:4\nSnippetByteRange:10:23",
            "Error while parsing Snippet: ['Multiple values for SnippetByteRange found. " "Line: 3']",
        ),
        (
            "SnippetSPDXID: SPDXRef-Snippet",
            r"__init__() missing 2 required " r"positional arguments: 'file_spdx_id' and 'byte_range'",
        ),
    ],
)
def test_parse_invalid_snippet(snippet_str, expected_message):
    parser = Parser()

    with pytest.raises(SPDXParsingError) as err:
        parser.parse(snippet_str)

    assert expected_message in err.value.get_messages()[0]



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import TextIO

from spdx_tools.spdx3.model import Annotation
from spdx_tools.spdx3.writer.console.console import write_value
from spdx_tools.spdx3.writer.console.element_writer import write_element_properties


def write_annotation(annotation: Annotation, text_output: TextIO):
    text_output.write("## Annotation\n")
    write_element_properties(annotation, text_output)

    for property_name in Annotation.__annotations__.keys():
        write_value(property_name, getattr(annotation, property_name), text_output)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import os
from unittest import TestCase

import pytest
from license_expression import get_spdx_licensing
from rdflib import RDF, BNode, Graph, URIRef

from spdx_tools.spdx.model import Checksum, ChecksumAlgorithm, FileType, SpdxNoAssertion
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.rdf.file_parser import parse_file
from spdx_tools.spdx.rdfschema.namespace import SPDX_NAMESPACE


def test_parse_file():
    graph = Graph().parse(os.path.join(os.path.dirname(__file__), "data/file_to_test_rdf_parser.rdf.xml"))
    file_node = graph.value(predicate=RDF.type, object=SPDX_NAMESPACE.File)
    doc_namespace = "https://some.namespace"
    assert isinstance(file_node, URIRef)

    file = parse_file(file_node, graph, doc_namespace)

    assert file.name == "./fileName.py"
    assert file.spdx_id == "SPDXRef-File"
    assert file.checksums == [Checksum(ChecksumAlgorithm.SHA1, "71c4025dd9897b364f3ebbb42c484ff43d00791c")]
    assert file.file_types == [FileType.TEXT]
    assert file.comment == "fileComment"
    assert file.copyright_text == "copyrightText"
    assert file.contributors == ["fileContributor"]
    assert file.license_concluded == get_spdx_licensing().parse("MIT AND GPL-2.0")
    TestCase().assertCountEqual(
        file.license_info_in_file,
        [get_spdx_licensing().parse("MIT"), get_spdx_licensing().parse("GPL-2.0"), SpdxNoAssertion()],
    )
    assert file.license_comment == "licenseComment"
    assert file.notice == "fileNotice"
    assert file.attribution_texts == ["fileAttributionText"]


def test_parse_invalid_file():
    graph = Graph().parse(os.path.join(os.path.dirname(__file__), "data/invalid_documents/file_without_spdx_ids.xml"))
    file_node = graph.value(predicate=RDF.type, object=SPDX_NAMESPACE.File)
    doc_namespace = "https://some.namespace"

    assert isinstance(file_node, BNode)
    with pytest.raises(SPDXParsingError):
        parse_file(file_node, graph, doc_namespace)



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import pytest

from spdx_tools.spdx.jsonschema.package_verification_code_converter import PackageVerificationCodeConverter
from spdx_tools.spdx.jsonschema.package_verification_code_properties import PackageVerificationCodeProperty
from spdx_tools.spdx.model import PackageVerificationCode


@pytest.fixture
def converter() -> PackageVerificationCodeConverter:
    return PackageVerificationCodeConverter()


@pytest.mark.parametrize(
    "package_verification_code_property,expected",
    [
        (
            PackageVerificationCodeProperty.PACKAGE_VERIFICATION_CODE_EXCLUDED_FILES,
            "packageVerificationCodeExcludedFiles",
        ),
        (PackageVerificationCodeProperty.PACKAGE_VERIFICATION_CODE_VALUE, "packageVerificationCodeValue"),
    ],
)
def test_json_property_names(
    converter: PackageVerificationCodeConverter,
    package_verification_code_property: PackageVerificationCodeProperty,
    expected: str,
):
    assert converter.json_property_name(package_verification_code_property) == expected


def test_json_type(converter: PackageVerificationCodeConverter):
    assert converter.get_json_type() == PackageVerificationCodeProperty


def test_data_model_type(converter: PackageVerificationCodeConverter):
    assert converter.get_data_model_type() == PackageVerificationCode


def test_successful_conversion(converter: PackageVerificationCodeConverter):
    package_verification_code = PackageVerificationCode("value", ["file1", "file2"])

    converted_dict = converter.convert(package_verification_code)

    assert converted_dict == {
        converter.json_property_name(PackageVerificationCodeProperty.PACKAGE_VERIFICATION_CODE_EXCLUDED_FILES): [
            "file1",
            "file2",
        ],
        converter.json_property_name(PackageVerificationCodeProperty.PACKAGE_VERIFICATION_CODE_VALUE): "value",
    }


def test_null_values(converter: PackageVerificationCodeConverter):
    package_verification_code = PackageVerificationCode("value")

    converted_dict = converter.convert(package_verification_code)

    assert (
        converter.json_property_name(PackageVerificationCodeProperty.PACKAGE_VERIFICATION_CODE_EXCLUDED_FILES)
        not in converted_dict
    )



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import Any, Type

from spdx_tools.spdx.datetime_conversions import datetime_to_iso_string
from spdx_tools.spdx.jsonschema.annotation_properties import AnnotationProperty
from spdx_tools.spdx.jsonschema.converter import TypedConverter
from spdx_tools.spdx.jsonschema.json_property import JsonProperty
from spdx_tools.spdx.model import Annotation, Document


class AnnotationConverter(TypedConverter[Annotation]):
    def _get_property_value(
        self, annotation: Annotation, annotation_property: AnnotationProperty, document: Document = None
    ) -> Any:
        if annotation_property == AnnotationProperty.ANNOTATION_DATE:
            return datetime_to_iso_string(annotation.annotation_date)
        elif annotation_property == AnnotationProperty.ANNOTATION_TYPE:
            return annotation.annotation_type.name
        elif annotation_property == AnnotationProperty.ANNOTATOR:
            return annotation.annotator.to_serialized_string()
        elif annotation_property == AnnotationProperty.COMMENT:
            return annotation.annotation_comment

    def get_json_type(self) -> Type[JsonProperty]:
        return AnnotationProperty

    def get_data_model_type(self) -> Type[Annotation]:
        return Annotation



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import pytest

from spdx_tools.spdx3.model import (
    Agent,
    Annotation,
    Bom,
    Bundle,
    LifecycleScopedRelationship,
    Organization,
    Person,
    Relationship,
    SoftwareAgent,
    SpdxDocument,
)
from spdx_tools.spdx3.model.ai import AIPackage
from spdx_tools.spdx3.model.build import Build
from spdx_tools.spdx3.model.dataset import Dataset
from spdx_tools.spdx3.model.licensing import (
    CustomLicense,
    CustomLicenseAddition,
    ListedLicense,
    ListedLicenseException,
)
from spdx_tools.spdx3.model.security import (
    CvssV2VulnAssessmentRelationship,
    CvssV3VulnAssessmentRelationship,
    EpssVulnAssessmentRelationship,
    ExploitCatalogVulnAssessmentRelationship,
    SsvcVulnAssessmentRelationship,
    VexAffectedVulnAssessmentRelationship,
    VexFixedVulnAssessmentRelationship,
    VexNotAffectedVulnAssessmentRelationship,
    VexUnderInvestigationVulnAssessmentRelationship,
    Vulnerability,
)
from spdx_tools.spdx3.model.software import File, Package, Sbom, Snippet, SoftwareDependencyRelationship
from tests.spdx3.fixtures import fixture_factory, get_fixture_dict
from tests.spdx3.model.model_test_utils import InvalidTypeClass, get_property_names

CLASS_LIST = [
    Agent,
    Person,
    Organization,
    SoftwareAgent,
    Annotation,
    Relationship,
    LifecycleScopedRelationship,
    Bundle,
    Bom,
    SpdxDocument,
    ListedLicense,
    CustomLicense,
    ListedLicenseException,
    CustomLicenseAddition,
    CvssV2VulnAssessmentRelationship,
    CvssV3VulnAssessmentRelationship,
    EpssVulnAssessmentRelationship,
    ExploitCatalogVulnAssessmentRelationship,
    SsvcVulnAssessmentRelationship,
    VexAffectedVulnAssessmentRelationship,
    VexFixedVulnAssessmentRelationship,
    VexNotAffectedVulnAssessmentRelationship,
    VexUnderInvestigationVulnAssessmentRelationship,
    Vulnerability,
    File,
    Package,
    Snippet,
    Sbom,
    SoftwareDependencyRelationship,
    Dataset,
    AIPackage,
    Build,
]


@pytest.mark.parametrize(
    "clazz",
    CLASS_LIST,
)
def test_correct_initialization(clazz):
    clazz_instance = fixture_factory(clazz)
    fixture_dict = get_fixture_dict(clazz)

    for property_name in get_property_names(clazz):
        assert getattr(clazz_instance, property_name) is not None
        assert getattr(clazz_instance, property_name) == fixture_dict[property_name]


@pytest.mark.parametrize(
    "clazz",
    CLASS_LIST,
)
def test_invalid_initialization(clazz):
    property_names = get_property_names(clazz)
    false_properties_dict = {}
    for property_name in property_names:
        false_properties_dict[property_name] = InvalidTypeClass()

    with pytest.raises(TypeError) as err:
        fixture_factory(clazz, **false_properties_dict)

    assert len(err.value.args[0]) == len(property_names)
    for error in err.value.args[0]:
        assert error.startswith(f"SetterError {clazz.__name__}")



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import List
from semantic_version import Version

from spdx_tools.spdx3.bump_from_spdx2.actor import bump_actor
from spdx_tools.spdx3.bump_from_spdx2.external_document_ref import bump_external_document_ref
from spdx_tools.spdx3.bump_from_spdx2.message import print_missing_conversion
from spdx_tools.spdx3.model import CreationInfo, ProfileIdentifier, SpdxDocument
from spdx_tools.spdx3.payload import Payload
from spdx_tools.spdx.model.actor import ActorType
from spdx_tools.spdx.model.document import CreationInfo as Spdx2_CreationInfo


def bump_creation_info(spdx2_creation_info: Spdx2_CreationInfo, payload: Payload) -> SpdxDocument:
    document_namespace = spdx2_creation_info.document_namespace
    spdx_id = f"{document_namespace}#{spdx2_creation_info.spdx_id}"

    print_missing_conversion("creation_info.document_namespace", 0, "https://github.com/spdx/spdx-3-model/issues/87")

    namespaces, imports = (
        zip(
            *[
                bump_external_document_ref(external_document_ref)
                for external_document_ref in spdx2_creation_info.external_document_refs
            ]
        )
        if spdx2_creation_info.external_document_refs
        else ([], [])
    )
    namespaces = list(namespaces)
    imports = list(imports)
    print_missing_conversion(
        "creation_info.license_list_version",
        0,
        "part of licensing profile, " "https://github.com/spdx/spdx-3-model/issues/131",
    )
    creation_info = CreationInfo(
        spec_version=Version("3.0.0"),
        created=spdx2_creation_info.created,
        created_by=[],
        profile=[ProfileIdentifier.CORE, ProfileIdentifier.SOFTWARE, ProfileIdentifier.LICENSING],
        data_license="https://spdx.org/licenses/" + spdx2_creation_info.data_license,
    )

    # due to creators having a creation_info themselves which inherits from the document's one,
    # we have to add them after the creation_info has been initialized
    creator_ids: List[str] = []
    tool_ids: List[str] = []
    for creator in spdx2_creation_info.creators:
        bumped_actor_id = bump_actor(creator, payload, document_namespace, creation_info)
        if creator.actor_type in [ActorType.PERSON, ActorType.ORGANIZATION]:
            creator_ids.append(bumped_actor_id)
        else:
            tool_ids.append(bumped_actor_id)

    if not creator_ids:
        print_missing_conversion(
            "Creators",
            0,
            "The SPDX2 creation_info does not contain creators of Type Person or Organization."
            " This case leads to an invalid SPDX3 document and is currently not supported."
            "https://github.com/spdx/spdx-3-model/issues/180",
        )

    creation_info.created_by = creator_ids
    creation_info.created_using = tool_ids

    return SpdxDocument(
        spdx_id=spdx_id,
        creation_info=creation_info,
        name=spdx2_creation_info.name,
        comment=spdx2_creation_info.document_comment,
        element=[],
        root_element=[],
        imports=imports,
        namespaces=namespaces,
    )



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import TextIO

from spdx_tools.spdx3.model import Agent, Organization, Person, SoftwareAgent
from spdx_tools.spdx3.writer.console.element_writer import write_element_properties


def write_agent(agent: Agent, text_output: TextIO, heading: bool = True):
    if heading:
        if isinstance(agent, Person):
            text_output.write("## Person\n")
        if isinstance(agent, Organization):
            text_output.write("## Organization\n")
        if isinstance(agent, SoftwareAgent):
            text_output.write("## SoftwareAgent\n")
    write_element_properties(agent, text_output)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import TextIO

from spdx_tools.spdx3.model.dataset import Dataset
from spdx_tools.spdx3.writer.console.console import write_value
from spdx_tools.spdx3.writer.console.software.package_writer import write_package


def write_dataset(dataset: Dataset, text_output: TextIO):
    text_output.write("## Dataset\n")
    write_package(dataset, text_output, False)

    for property_name in Dataset.__annotations__.keys():
        write_value(property_name, getattr(dataset, property_name), text_output)



#  SPDX-FileCopyrightText: 2023 spdx contributors
#
#  SPDX-License-Identifier: Apache-2.0
from os import path

from spdx_tools.spdx.model import Document
from spdx_tools.spdx3.payload import Payload
from spdx_tools.spdx3.writer.json_ld.json_ld_writer import write_payload
from spdx_tools.spdx3.bump_from_spdx2.spdx_document import bump_spdx_document
from spdx_tools.spdx.parser.parse_anything import parse_file

# This example demonstrates how to load an existing SPDX2 file and convert it to the SPDX3 format

# Provide a path to the input file
input_path = path.join(path.dirname(__file__), "..", "tests", "spdx", "data", "SPDXLite.spdx")
# Parse the original SPDX2 input file
spdx2_document: Document = parse_file(input_path)
# Convert original document to an SPDX3 payload
spdx3_payload: Payload = bump_spdx_document(spdx2_document)
# Write SPDX3 payload in json-ld format
write_payload(spdx3_payload, "spdx2_to_3")



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

import sys

MISSING_CONVERSION_REASONS = {0: "missing conversion rule", 1: "missing implementation"}


def print_missing_conversion(field: str, reason, additional_information: str = ""):
    print(f"{field} not converted: {MISSING_CONVERSION_REASONS[reason]} {additional_information}", file=sys.stderr)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import TextIO

from spdx_tools.spdx3.model.software import Package
from spdx_tools.spdx3.writer.console.artifact_writer import write_artifact_properties
from spdx_tools.spdx3.writer.console.console import write_value


def write_package(package: Package, text_output: TextIO, heading: bool = True):
    if heading:
        text_output.write("## Package\n")
    write_artifact_properties(package, text_output)

    for property_name in Package.__annotations__.keys():
        if property_name == "package_purpose":
            write_value(
                property_name, ", ".join([purpose.name for purpose in getattr(package, property_name)]), text_output
            )
            continue
        write_value(property_name, getattr(package, property_name), text_output)



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from unittest import TestCase

import pytest

from spdx_tools.spdx.model import ActorType
from spdx_tools.spdx.parser.actor_parser import ActorParser
from spdx_tools.spdx.parser.error import SPDXParsingError


@pytest.mark.parametrize(
    "actor_string,expected_type,expected_name,expected_mail",
    [
        ("Person: Jane Doe (jane.doe@example.com)", ActorType.PERSON, "Jane Doe", "jane.doe@example.com"),
        (
            "Organization: Example organization (organization@example.com)",
            ActorType.ORGANIZATION,
            "Example organization",
            "organization@example.com",
        ),
        ("Organization: Example organization ( )", ActorType.ORGANIZATION, "Example organization", None),
        ("Tool: Example tool ", ActorType.TOOL, "Example tool", None),
    ],
)
def test_parse_actor(actor_string, expected_type, expected_name, expected_mail):
    actor_parser = ActorParser()

    actor = actor_parser.parse_actor(actor_string)

    assert actor.actor_type == expected_type
    assert actor.name == expected_name
    assert actor.email == expected_mail


@pytest.mark.parametrize(
    "actor_string,expected_message",
    [
        (
            "Perso: Jane Doe (jane.doe@example.com)",
            ["Actor Perso: Jane Doe (jane.doe@example.com) doesn't match any of person, organization or tool."],
        ),
        ("Toole Example Tool ()", ["Actor Toole Example Tool () doesn't match any of person, organization or tool."]),
    ],
)
def test_parse_invalid_actor(actor_string, expected_message):
    actor_parser = ActorParser()
    actor_string = actor_string

    with pytest.raises(SPDXParsingError) as err:
        actor_parser.parse_actor(actor_string)

    TestCase().assertCountEqual(err.value.get_messages(), expected_message)



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import re

from beartype.typing import Match, Optional, Pattern

from spdx_tools.spdx.model import Actor, ActorType
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.parsing_functions import construct_or_raise_parsing_error


class ActorParser:
    @staticmethod
    def parse_actor(actor: str) -> Actor:
        tool_re: Pattern = re.compile(r"^Tool:\s*(.+)", re.UNICODE)
        person_re: Pattern = re.compile(r"^Person:\s*(([^(])+)(\((.*)\))?", re.UNICODE)
        org_re: Pattern = re.compile(r"^Organization:\s*(([^(])+)(\((.*)\))?", re.UNICODE)
        tool_match: Match = tool_re.match(actor)
        person_match: Match = person_re.match(actor)
        org_match: Match = org_re.match(actor)

        if tool_match:
            name: str = tool_match.group(1).strip()
            if not name:
                raise SPDXParsingError([f"No name for Tool provided: {actor}."])
            creator = construct_or_raise_parsing_error(Actor, dict(actor_type=ActorType.TOOL, name=name))

        elif person_match:
            name: str = person_match.group(1).strip()
            if not name:
                raise SPDXParsingError([f"No name for Person provided: {actor}."])
            email: Optional[str] = ActorParser.get_email_or_none(person_match)
            creator = construct_or_raise_parsing_error(
                Actor, dict(actor_type=ActorType.PERSON, name=name, email=email)
            )
        elif org_match:
            name: str = org_match.group(1).strip()
            if not name:
                raise SPDXParsingError([f"No name for Organization provided: {actor}."])
            email: Optional[str] = ActorParser.get_email_or_none(org_match)
            creator = construct_or_raise_parsing_error(
                Actor, dict(actor_type=ActorType.ORGANIZATION, name=name, email=email)
            )
        else:
            raise SPDXParsingError([f"Actor {actor} doesn't match any of person, organization or tool."])

        return creator

    @staticmethod
    def get_email_or_none(match: Match) -> Optional[str]:
        email_match = match.group(4)
        if email_match and email_match.strip():
            email = email_match.strip()
        else:
            email = None
        return email



#!/usr/bin/env python3

# Copyright (c) 2020 Yash Varshney
# Copyright (c) 2023 spdx contributors
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#    http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import sys

import click
from beartype.typing import List

from spdx_tools.spdx.graph_generation import export_graph_from_document
from spdx_tools.spdx.model import Document
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.parse_anything import parse_file
from spdx_tools.spdx.validation.document_validator import validate_full_spdx_document
from spdx_tools.spdx.validation.validation_message import ValidationMessage
from spdx_tools.spdx.writer.tagvalue import tagvalue_writer
from spdx_tools.spdx.writer.write_anything import write_file


@click.command()
@click.option("--infile", "-i", required=True, help="The file containing the document to be validated or converted.")
@click.option(
    "--outfile",
    "-o",
    help="The file to write the converted document to (write a dash for output to stdout or omit for no conversion). "
    "If you add the option --graph to the command the generated graph will be written to this file.",
)
@click.option(
    "--version",
    help='The SPDX version to be used during parsing and validation ("SPDX-2.2" or "SPDX-2.3"). '
    "Will be read from the document if not provided.",
    default=None,
)
@click.option("--novalidation", is_flag=True, help="Don't validate the provided document.")
@click.option(
    "--graph",
    is_flag=True,
    default=False,
    help="Generate a relationship graph from the input file. "
    "The generated graph is saved to the file specified with --outfile. "
    "Note: You need to install the optional dependencies 'networkx' and 'pygraphviz' for this feature.",
)
def main(infile: str, outfile: str, version: str, novalidation: bool, graph: bool):
    """
    CLI-tool for validating SPDX documents and converting between RDF, TAG-VALUE, JSON, YAML and XML formats.
    Formats are determined by the file endings.
    To use, run: 'pyspdxtools --infile <input file name> --outfile <output file name>'
    """
    try:
        document: Document = parse_file(infile)

        if not novalidation:
            if not version:
                version = document.creation_info.spdx_version

            if version not in ["SPDX-2.2", "SPDX-2.3"]:
                logging.error(f"This tool only supports SPDX versions SPDX-2.2 and SPDX-2.3, but got: {version}")
                sys.exit(1)

            validation_messages: List[ValidationMessage] = validate_full_spdx_document(document, version)
            if validation_messages:
                log_string = "\n".join(
                    ["The document is invalid. The following issues have been found:"]
                    + [message.validation_message for message in validation_messages]
                )
                logging.error(log_string)
                sys.exit(1)
            else:
                logging.info("The document is valid.")

        if outfile == "-":
            tagvalue_writer.write_document(document, sys.stdout)

        elif graph:
            try:
                export_graph_from_document(document, outfile)
            except ImportError:
                logging.error(
                    "To be able to draw a relationship graph of the parsed document "
                    "you need to install 'networkx' and 'pygraphviz'. Run 'pip install \".[graph_generation]\"'."
                )
                sys.exit(1)

        elif outfile:
            write_file(document, outfile, validate=False)

    except NotImplementedError as err:
        logging.error(
            err.args[0]
            + "\nPlease note that this project is currently undergoing a major refactoring and therefore missing "
            "a few features which will be added in time (refer to https://github.com/spdx/tools-python/issues "
            "for insights into the current status).\n"
            "In the meantime, please use the current PyPI release version."
        )
        sys.exit(1)

    except SPDXParsingError as err:
        log_string = "\n".join(
            ["There have been issues while parsing the provided document:"]
            + [message for message in err.get_messages()]
        )
        logging.error(log_string)
        sys.exit(1)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import TextIO

from spdx_tools.spdx3.model import Element
from spdx_tools.spdx3.writer.console.console import write_value
from spdx_tools.spdx3.writer.console.creation_info_writer import write_creation_info
from spdx_tools.spdx3.writer.console.external_identifier_writer import write_external_identifier
from spdx_tools.spdx3.writer.console.external_reference_writer import write_external_reference
from spdx_tools.spdx3.writer.console.hash_writer import write_hash
from spdx_tools.spdx.writer.tagvalue.tagvalue_writer_helper_functions import write_optional_heading


def write_element_properties(element: Element, text_output: TextIO):
    write_value("SPDXID", element.spdx_id, text_output)
    write_value("name", element.name, text_output)
    if element.creation_info:
        write_creation_info(element.creation_info, text_output, True)
    write_value("summary", element.summary, text_output)
    write_value("description", element.description, text_output)
    write_value("comment", element.comment, text_output)
    write_optional_heading(element.verified_using, "verified using:\n", text_output)
    for integrity_method in element.verified_using:
        # for now Hash is the only child class of the abstract class IntegrityMethod,
        # as soon as there are more inherited classes we need to implement a logic
        # that determines the correct write function for the "integrity_method" object
        write_hash(integrity_method, text_output, heading=False)
    write_optional_heading(element.external_reference, "External Reference\n", text_output)
    for external_reference in element.external_reference:
        write_external_reference(external_reference, text_output)
    write_optional_heading(element.external_identifier, "External Identifier\n", text_output)
    for external_identifier in element.external_identifier:
        write_external_identifier(external_identifier, text_output)



# Copyright (c) 2014 Ahmed H. Ismail
# Copyright (c) 2023 spdx contributors
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ply import lex
from ply.lex import TOKEN


class SPDXLexer:
    reserved = {
        # Top level fields
        "SPDXVersion": "DOC_VERSION",
        "DataLicense": "DOC_LICENSE",
        "DocumentName": "DOC_NAME",
        "SPDXID": "SPDX_ID",
        "DocumentComment": "DOC_COMMENT",
        "DocumentNamespace": "DOC_NAMESPACE",
        "ExternalDocumentRef": "EXT_DOC_REF",
        # Creation info fields
        "Creator": "CREATOR",
        "Created": "CREATED",
        "CreatorComment": "CREATOR_COMMENT",
        "LicenseListVersion": "LICENSE_LIST_VERSION",
        # Annotation fields
        "Annotator": "ANNOTATOR",
        "AnnotationDate": "ANNOTATION_DATE",
        "AnnotationComment": "ANNOTATION_COMMENT",
        "AnnotationType": "ANNOTATION_TYPE",
        "SPDXREF": "ANNOTATION_SPDX_ID",
        # Relationship fields
        "Relationship": "RELATIONSHIP",
        "RelationshipComment": "RELATIONSHIP_COMMENT",
        # Package fields
        "PackageName": "PKG_NAME",
        "PackageVersion": "PKG_VERSION",
        "PackageDownloadLocation": "PKG_DOWNLOAD_LOCATION",
        "FilesAnalyzed": "PKG_FILES_ANALYZED",
        "PackageSummary": "PKG_SUMMARY",
        "PackageSourceInfo": "PKG_SOURCE_INFO",
        "PackageFileName": "PKG_FILE_NAME",
        "PackageSupplier": "PKG_SUPPLIER",
        "PackageOriginator": "PKG_ORIGINATOR",
        "PackageChecksum": "PKG_CHECKSUM",
        "PackageVerificationCode": "PKG_VERIFICATION_CODE",
        "PackageDescription": "PKG_DESCRIPTION",
        "PackageComment": "PKG_COMMENT",
        "PackageLicenseDeclared": "PKG_LICENSE_DECLARED",
        "PackageLicenseConcluded": "PKG_LICENSE_CONCLUDED",
        "PackageLicenseInfoFromFiles": "PKG_LICENSE_INFO",
        "PackageLicenseComments": "PKG_LICENSE_COMMENT",
        "PackageCopyrightText": "PKG_COPYRIGHT_TEXT",
        "PackageHomePage": "PKG_HOMEPAGE",
        "ExternalRef": "PKG_EXTERNAL_REF",
        "ExternalRefComment": "PKG_EXTERNAL_REF_COMMENT",
        "PackageAttributionText": "PKG_ATTRIBUTION_TEXT",
        "PrimaryPackagePurpose": "PRIMARY_PACKAGE_PURPOSE",
        "BuiltDate": "BUILT_DATE",
        "ReleaseDate": "RELEASE_DATE",
        "ValidUntilDate": "VALID_UNTIL_DATE",
        # File fields
        "FileName": "FILE_NAME",
        "FileType": "FILE_TYPE",
        "FileChecksum": "FILE_CHECKSUM",
        "LicenseConcluded": "FILE_LICENSE_CONCLUDED",
        "LicenseInfoInFile": "FILE_LICENSE_INFO",
        "FileCopyrightText": "FILE_COPYRIGHT_TEXT",
        "LicenseComments": "FILE_LICENSE_COMMENT",
        "FileComment": "FILE_COMMENT",
        "FileNotice": "FILE_NOTICE",
        "FileContributor": "FILE_CONTRIBUTOR",
        "FileAttributionText": "FILE_ATTRIBUTION_TEXT",
        # ExtractedLicensingInfo fields
        "LicenseID": "LICENSE_ID",
        "ExtractedText": "LICENSE_TEXT",
        "LicenseName": "LICENSE_NAME",
        "LicenseCrossReference": "LICENSE_CROSS_REF",
        "LicenseComment": "LICENSE_COMMENT",
        # Snippet fields
        "SnippetSPDXID": "SNIPPET_SPDX_ID",
        "SnippetName": "SNIPPET_NAME",
        "SnippetComment": "SNIPPET_COMMENT",
        "SnippetCopyrightText": "SNIPPET_COPYRIGHT_TEXT",
        "SnippetLicenseComments": "SNIPPET_LICENSE_COMMENT",
        "SnippetFromFileSPDXID": "SNIPPET_FILE_SPDXID",
        "SnippetLicenseConcluded": "SNIPPET_LICENSE_CONCLUDED",
        "LicenseInfoInSnippet": "SNIPPET_LICENSE_INFO",
        "SnippetAttributionText": "SNIPPET_ATTRIBUTION_TEXT",
        "SnippetByteRange": "SNIPPET_BYTE_RANGE",
        "SnippetLineRange": "SNIPPET_LINE_RANGE",
        # Common fields
        "NOASSERTION": "NO_ASSERTION",
        "NONE": "NONE",
    }
    states = (("text", "exclusive"),)

    tokens = [
        "TEXT",
        "TOOL_VALUE",
        "UNKNOWN_TAG",
        "ORGANIZATION_VALUE",
        "PERSON_VALUE",
        "ISO8601_DATE",
        "LINE",
        "CHECKSUM",
    ] + list(reserved.values())

    def __init__(self):
        self.lexer = None

    @TOKEN(r":\s*<text>")
    def t_text(self, t):
        t.lexer.text_start = t.lexer.lexpos - len("<text>")
        t.lexer.begin("text")

    @TOKEN(r"</text>\s*")
    def t_text_end(self, t):
        t.type = "TEXT"
        t.value = t.lexer.lexdata[t.lexer.text_start : t.lexer.lexpos]
        t.lexer.lineno += t.value.count("\n")
        t.value = t.value.strip()
        t.lexer.begin("INITIAL")
        return t

    @TOKEN(r".|\n")
    def t_text_any(self, t):
        pass

    def t_text_error(self, t):
        print("Lexer error in text state")

    @TOKEN(
        r":\s*(ADLER32|BLAKE2b-256|BLAKE2b-384|BLAKE2b-512|BLAKE3|MD2|MD4|MD5|MD6|SHA1|SHA224|SHA256|SHA384|SHA512|"
        r"SHA3-256|SHA3-384|SHA3-512):\s*([a-f0-9]*)"
    )
    def t_CHECKSUM(self, t):
        t.value = t.value[1:].strip()
        return t

    @TOKEN(r":\s*Tool:.+")
    def t_TOOL_VALUE(self, t):
        t.value = t.value[1:].strip()
        return t

    @TOKEN(r":\s*Organization:.+")
    def t_ORGANIZATION_VALUE(self, t):
        t.value = t.value[1:].strip()
        return t

    @TOKEN(r":\s*Person:.+")
    def t_PERSON_VALUE(self, t):
        t.value = t.value[1:].strip()
        return t

    @TOKEN(r":\s*\d\d\d\d-\d\d-\d\dT\d\d:\d\d:\d\dZ")
    def t_ISO8601_DATE(self, t):
        t.value = t.value[1:].strip()
        return t

    @TOKEN(r"[a-zA-Z]+")
    def t_KEYWORD_AS_TAG(self, t):
        t.type = self.reserved.get(t.value, "UNKNOWN_TAG")
        t.value = t.value.strip()
        return t

    @TOKEN(r":.+")
    def t_LINE_OR_KEYWORD_VALUE(self, t):
        t.value = t.value[1:].strip()
        if t.value in self.reserved.keys():
            t.type = self.reserved[t.value]
        else:
            t.type = "LINE"
        return t

    @TOKEN(r"\#.*")
    def t_comment(self, t):
        pass

    @TOKEN(r"\n+")
    def t_newline(self, t):
        t.lexer.lineno += len(t.value)

    @TOKEN(r"[ \t]+")
    def t_whitespace(self, t):
        pass

    def build(self, **kwargs):
        self.lexer = lex.lex(module=self, **kwargs)

    def token(self):
        return self.lexer.token()

    def input(self, data):
        self.lexer.input(data)

    def t_error(self, t):
        t.lexer.skip(1)
        t.value = "Lexer error"
        return t



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import List, Optional

from spdx_tools.common.typing.dataclass_with_properties import dataclass_with_properties
from spdx_tools.common.typing.type_checks import check_types_and_set_values
from spdx_tools.spdx3.model.licensing.license_addition import LicenseAddition


@dataclass_with_properties
class CustomLicenseAddition(LicenseAddition):
    def __init__(
        self,
        addition_id: str,
        addition_name: str,
        addition_text: str,
        addition_comment: Optional[str] = None,
        see_also: List[str] = None,
        standard_addition_template: Optional[str] = None,
        is_deprecated_addition_id: Optional[bool] = None,
        obsoleted_by: Optional[str] = None,
    ):
        see_also = [] if see_also is None else see_also
        check_types_and_set_values(self, locals())



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from rdflib import RDF, RDFS, BNode, Graph, Literal, URIRef

from spdx_tools.spdx.model import ExtractedLicensingInfo
from spdx_tools.spdx.rdfschema.namespace import SPDX_NAMESPACE
from spdx_tools.spdx.writer.rdf.writer_utils import add_literal_or_no_assertion, add_optional_literal


def add_extracted_licensing_info_to_graph(
    extracted_licensing_info: ExtractedLicensingInfo, graph: Graph, doc_node, doc_namespace: str
):
    if extracted_licensing_info.license_id:
        extracted_licensing_info_resource = URIRef(f"{doc_namespace}#{extracted_licensing_info.license_id}")
        graph.add((extracted_licensing_info_resource, RDF.type, SPDX_NAMESPACE.ExtractedLicensingInfo))
    else:
        extracted_licensing_info_resource = BNode()
    add_optional_literal(
        extracted_licensing_info.license_id, graph, extracted_licensing_info_resource, SPDX_NAMESPACE.licenseId
    )
    add_optional_literal(
        extracted_licensing_info.extracted_text, graph, extracted_licensing_info_resource, SPDX_NAMESPACE.extractedText
    )
    add_literal_or_no_assertion(
        extracted_licensing_info.license_name, graph, extracted_licensing_info_resource, SPDX_NAMESPACE.name
    )
    for cross_reference in extracted_licensing_info.cross_references:
        graph.add((extracted_licensing_info_resource, RDFS.seeAlso, Literal(cross_reference)))
    add_optional_literal(extracted_licensing_info.comment, graph, extracted_licensing_info_resource, RDFS.comment)

    graph.add((doc_node, SPDX_NAMESPACE.hasExtractedLicensingInfo, extracted_licensing_info_resource))



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import Dict
from rdflib import RDF, RDFS, Graph, Literal, URIRef

from spdx_tools.spdx.casing_tools import snake_case_to_camel_case
from spdx_tools.spdx.model import File
from spdx_tools.spdx.rdfschema.namespace import SPDX_NAMESPACE
from spdx_tools.spdx.writer.rdf.checksum_writer import add_checksum_to_graph
from spdx_tools.spdx.writer.rdf.license_expression_writer import add_license_expression_or_none_or_no_assertion
from spdx_tools.spdx.writer.rdf.writer_utils import add_namespace_to_spdx_id, add_optional_literal


def add_file_to_graph(file: File, graph: Graph, doc_namespace: str, external_doc_ref_to_namespace: Dict[str, str]):
    file_resource = URIRef(add_namespace_to_spdx_id(file.spdx_id, doc_namespace, external_doc_ref_to_namespace))
    graph.add((file_resource, RDF.type, SPDX_NAMESPACE.File))
    graph.add((file_resource, SPDX_NAMESPACE.fileName, Literal(file.name)))
    for file_type in file.file_types:
        graph.add(
            (
                file_resource,
                SPDX_NAMESPACE.fileType,
                SPDX_NAMESPACE[f"fileType_{snake_case_to_camel_case(file_type.name)}"],
            )
        )

    for checksum in file.checksums:
        add_checksum_to_graph(checksum, graph, file_resource)

    add_license_expression_or_none_or_no_assertion(
        file.license_concluded, graph, file_resource, SPDX_NAMESPACE.licenseConcluded, doc_namespace
    )
    add_license_expression_or_none_or_no_assertion(
        file.license_info_in_file, graph, file_resource, SPDX_NAMESPACE.licenseInfoInFile, doc_namespace
    )

    add_optional_literal(file.license_comment, graph, file_resource, SPDX_NAMESPACE.licenseComments)
    add_optional_literal(file.copyright_text, graph, file_resource, SPDX_NAMESPACE.copyrightText)
    add_optional_literal(file.comment, graph, file_resource, RDFS.comment)
    add_optional_literal(file.notice, graph, file_resource, SPDX_NAMESPACE.noticeText)
    for contributor in file.contributors:
        graph.add((file_resource, SPDX_NAMESPACE.fileContributor, Literal(contributor)))
    for attribution_text in file.attribution_texts:
        graph.add((file_resource, SPDX_NAMESPACE.attributionText, Literal(attribution_text)))



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from unittest import TestCase

import pytest

from spdx_tools.spdx.constants import DOCUMENT_SPDX_ID
from spdx_tools.spdx.validation.spdx_id_validators import (
    get_list_of_all_spdx_ids,
    is_external_doc_ref_present_in_document,
    is_spdx_id_present_in_document,
    is_valid_external_doc_ref_id,
    is_valid_internal_spdx_id,
    validate_spdx_id,
)
from tests.spdx.fixtures import (
    creation_info_fixture,
    document_fixture,
    external_document_ref_fixture,
    file_fixture,
    package_fixture,
    snippet_fixture,
)

DOCUMENT = document_fixture(
    files=[file_fixture(spdx_id="SPDXRef-File1"), file_fixture(spdx_id="SPDXRef-File2")],
    packages=[package_fixture(spdx_id="SPDXRef-Package1"), package_fixture(spdx_id="SPDXRef-Package2")],
    snippets=[snippet_fixture(spdx_id="SPDXRef-Snippet1"), snippet_fixture(spdx_id="SPDXRef-Snippet2")],
    creation_info=creation_info_fixture(
        external_document_refs=[
            external_document_ref_fixture(document_ref_id="DocumentRef-external"),
            external_document_ref_fixture(document_ref_id="DocumentRef-1.2-ext"),
        ]
    ),
)


@pytest.mark.parametrize("spdx_id", [DOCUMENT_SPDX_ID, "SPDXRef-File1", "SPDXRef-1.3-3.7"])
def test_valid_internal_spdx_ids(spdx_id):
    assert is_valid_internal_spdx_id(spdx_id)


@pytest.mark.parametrize(
    "spdx_id", ["spdxId", "spdxRef-DOCUMENT", "SPDXRef.File", "SPDXRef#Snippet", "SPDXRef-1.3_3.7"]
)
def test_invalid_internal_spdx_ids(spdx_id):
    assert not is_valid_internal_spdx_id(spdx_id)


@pytest.mark.parametrize("doc_ref_id", ["DocumentRef-external", "DocumentRef-...+", "DocumentRef-v0.4.2-alpha"])
def test_valid_external_doc_ref_ids(doc_ref_id):
    assert is_valid_external_doc_ref_id(doc_ref_id)


@pytest.mark.parametrize(
    "doc_ref_id", ["external-ref", "Documentref-external", "DocumentRef-...#", "DocumentRef-v0_4_2-alpha"]
)
def test_invalid_external_doc_ref_ids(doc_ref_id):
    assert not is_valid_external_doc_ref_id(doc_ref_id)


def test_is_spdx_id_present_in_document():
    assert is_spdx_id_present_in_document("SPDXRef-File1", DOCUMENT)
    assert is_spdx_id_present_in_document("SPDXRef-Package2", DOCUMENT)
    assert is_spdx_id_present_in_document("SPDXRef-Snippet1", DOCUMENT)
    assert is_spdx_id_present_in_document(DOCUMENT_SPDX_ID, DOCUMENT)
    assert not is_spdx_id_present_in_document("SPDXRef-file2", DOCUMENT)


def test_is_external_doc_ref_present_in_document():
    assert is_external_doc_ref_present_in_document("DocumentRef-1.2-ext", DOCUMENT)
    assert not is_external_doc_ref_present_in_document("DocumentRef-External1", DOCUMENT)


def test_list_of_all_spdx_ids():
    TestCase().assertCountEqual(
        get_list_of_all_spdx_ids(DOCUMENT),
        [
            DOCUMENT_SPDX_ID,
            "SPDXRef-File1",
            "SPDXRef-File2",
            "SPDXRef-Package1",
            "SPDXRef-Package2",
            "SPDXRef-Snippet1",
            "SPDXRef-Snippet2",
        ],
    )


@pytest.mark.parametrize("spdx_id", ["DocumentRef-external:SPDXRef-File", "SPDXRef-Package"])
def test_valid_spdx_id(spdx_id):
    validation_messages = validate_spdx_id(spdx_id, DOCUMENT)

    assert validation_messages == []


@pytest.mark.parametrize(
    "spdx_id, expected_messages",
    [
        (
            "DocumentRef-external:extern:SPDXRef-File",
            [
                "spdx_id must not contain more than one colon in order to separate the external document reference id"
                " from the internal SPDX id, but is: DocumentRef-external:extern:SPDXRef-File"
            ],
        ),
        (
            "DocumentRef external:SPDXRef-File",
            [
                'the external document reference part of spdx_id must only contain letters, numbers, ".", "-" and "+" '
                'and must begin with "DocumentRef-", but is: DocumentRef external',
                'did not find the external document reference "DocumentRef external" in the SPDX document',
            ],
        ),
        (
            "DocRef-ext:SPDXRef-File_2",
            [
                'the external document reference part of spdx_id must only contain letters, numbers, ".", "-" and "+" '
                'and must begin with "DocumentRef-", but is: DocRef-ext',
                'the internal SPDX id part of spdx_id must only contain letters, numbers, "." and "-" and must begin '
                'with "SPDXRef-", but is: SPDXRef-File_2',
                'did not find the external document reference "DocRef-ext" in the SPDX document',
            ],
        ),
        (
            "DocumentRef-external:SPDXRef-File_2",
            [
                'the internal SPDX id part of spdx_id must only contain letters, numbers, "." and "-" and must begin '
                'with "SPDXRef-", but is: SPDXRef-File_2'
            ],
        ),
        (
            "SPDXRef-42+",
            [
                'spdx_id must only contain letters, numbers, "." and "-" and must begin with "SPDXRef-", but is: '
                "SPDXRef-42+"
            ],
        ),
    ],
)
def test_invalid_spdx_id(spdx_id, expected_messages):
    validation_messages = validate_spdx_id(spdx_id, DOCUMENT)

    TestCase().assertCountEqual(validation_messages, expected_messages)


@pytest.mark.parametrize(
    "spdx_id",
    ["DocumentRef-external:SPDXRef-File", DOCUMENT_SPDX_ID, "SPDXRef-File1", "SPDXRef-Package1", "SPDXRef-Snippet1"],
)
def test_valid_spdx_id_with_check_document(spdx_id):
    validation_messages = validate_spdx_id(spdx_id, DOCUMENT, check_document=True)
    assert validation_messages == []


def test_invalid_spdx_id_with_check_document():
    validation_messages = validate_spdx_id("SPDXRef-Filet", DOCUMENT, check_document=True)
    assert validation_messages == ['did not find the referenced spdx_id "SPDXRef-Filet" in the SPDX document']


@pytest.mark.parametrize("spdx_id", ["DocumentRef-external:SPDXRef-File", "SPDXRef-File1"])
def test_valid_spdx_id_with_check_files(spdx_id):
    validation_messages = validate_spdx_id(spdx_id, DOCUMENT, check_files=True)
    assert validation_messages == []


def test_invalid_spdx_id_with_check_files():
    validation_messages = validate_spdx_id("SPDXRef-Package1", DOCUMENT, check_files=True)
    assert validation_messages == [
        'did not find the referenced spdx_id "SPDXRef-Package1" in the SPDX document\'s files'
    ]



# SPDX-License-Identifier: Apache-2.0
#  Copyright (c) 2022 spdx contributors
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#    http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from beartype.typing import TextIO

from spdx_tools.spdx.model import ExtractedLicensingInfo
from spdx_tools.spdx.writer.tagvalue.tagvalue_writer_helper_functions import write_text_value, write_value


def write_extracted_licensing_info(extracted_licensing_info: ExtractedLicensingInfo, text_output: TextIO):
    write_value("LicenseID", extracted_licensing_info.license_id, text_output)
    write_text_value("ExtractedText", extracted_licensing_info.extracted_text, text_output)
    write_value("LicenseName", extracted_licensing_info.license_name, text_output)

    for cross_reference in sorted(extracted_licensing_info.cross_references):
        write_value("LicenseCrossReference", cross_reference, text_output)

    write_text_value("LicenseComment", extracted_licensing_info.comment, text_output)



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import Any, Type

from spdx_tools.spdx.datetime_conversions import datetime_to_iso_string
from spdx_tools.spdx.jsonschema.converter import TypedConverter
from spdx_tools.spdx.jsonschema.creation_info_properties import CreationInfoProperty
from spdx_tools.spdx.jsonschema.json_property import JsonProperty
from spdx_tools.spdx.jsonschema.optional_utils import apply_if_present
from spdx_tools.spdx.model import CreationInfo, Document


class CreationInfoConverter(TypedConverter[CreationInfo]):
    def get_data_model_type(self) -> Type[CreationInfo]:
        return CreationInfo

    def get_json_type(self) -> Type[JsonProperty]:
        return CreationInfoProperty

    def _get_property_value(
        self, creation_info: CreationInfo, creation_info_property: CreationInfoProperty, _document: Document = None
    ) -> Any:
        if creation_info_property == CreationInfoProperty.CREATED:
            return datetime_to_iso_string(creation_info.created)
        elif creation_info_property == CreationInfoProperty.CREATORS:
            return [creator.to_serialized_string() for creator in creation_info.creators] or None
        elif creation_info_property == CreationInfoProperty.LICENSE_LIST_VERSION:
            return apply_if_present(str, creation_info.license_list_version)
        elif creation_info_property == CreationInfoProperty.COMMENT:
            return creation_info.creator_comment



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from datetime import datetime

import pytest

from spdx_tools.spdx.datetime_conversions import datetime_from_str, datetime_to_iso_string


def test_datetime_to_iso_string():
    assert datetime_to_iso_string(datetime(2022, 12, 13, 1, 2, 3)) == "2022-12-13T01:02:03Z"


def test_datetime_to_iso_string_with_microseconds():
    assert datetime_to_iso_string(datetime(2022, 12, 13, 1, 2, 3, 666666)) == "2022-12-13T01:02:03Z"


def test_datetime_from_str():
    date_str = "2010-03-04T05:45:11Z"

    date = datetime_from_str(date_str)

    assert date == datetime(2010, 3, 4, 5, 45, 11)


@pytest.mark.parametrize(
    "invalid_date_str, error_type, expected_message",
    [
        (5, TypeError, "Could not convert str to datetime, invalid type: int"),
        ("2010-02-03", ValueError, "time data '2010-02-03' does not match format '%Y-%m-%dT%H:%M:%SZ'"),
    ],
)
def test_datetime_from_str_error(invalid_date_str, error_type, expected_message):
    with pytest.raises(error_type, match=expected_message):
        datetime_from_str(invalid_date_str)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import List, Union
from license_expression import (
    AND,
    OR,
    LicenseExpression,
    LicenseSymbol,
    LicenseWithExceptionSymbol,
    get_spdx_licensing,
)

from spdx_tools.spdx3.model.licensing import (
    AnyLicenseInfo,
    ConjunctiveLicenseSet,
    CustomLicense,
    CustomLicenseAddition,
    DisjunctiveLicenseSet,
    License,
    LicenseAddition,
    LicenseField,
    ListedLicense,
    ListedLicenseException,
    NoAssertionLicense,
    NoneLicense,
    WithAdditionOperator,
)
from spdx_tools.spdx.model import ExtractedLicensingInfo, SpdxNoAssertion, SpdxNone


def bump_license_expression_or_none_or_no_assertion(
    element: Union[LicenseExpression, SpdxNoAssertion, SpdxNone],
    extracted_licensing_info: List[ExtractedLicensingInfo],
) -> LicenseField:
    if isinstance(element, SpdxNone):
        return NoneLicense()
    elif isinstance(element, SpdxNoAssertion):
        return NoAssertionLicense()
    else:
        return bump_license_expression(element, extracted_licensing_info)


def bump_license_expression(
    license_expression: LicenseExpression, extracted_licensing_info: List[ExtractedLicensingInfo]
) -> AnyLicenseInfo:
    if isinstance(license_expression, AND):
        return ConjunctiveLicenseSet(
            member=[bump_license_expression(element, extracted_licensing_info) for element in license_expression.args]
        )
    if isinstance(license_expression, OR):
        return DisjunctiveLicenseSet(
            member=[bump_license_expression(element, extracted_licensing_info) for element in license_expression.args]
        )
    if isinstance(license_expression, LicenseWithExceptionSymbol):
        subject_license = bump_license_expression(license_expression.license_symbol, extracted_licensing_info)
        if not isinstance(subject_license, License):
            raise ValueError("Subject of LicenseException couldn't be converted to License.")
        return WithAdditionOperator(
            subject_license=subject_license,
            subject_addition=bump_license_exception(license_expression.exception_symbol, extracted_licensing_info),
        )
    if isinstance(license_expression, LicenseSymbol):
        if not get_spdx_licensing().validate(license_expression).invalid_symbols:
            return ListedLicense(license_expression.key, license_expression.obj, "blank")
        else:
            for licensing_info in extracted_licensing_info:
                if licensing_info.license_id == license_expression.key:
                    # the fields are optional in ExtractedLicensingInfo, to prevent type errors we use a type
                    # conversion to str as a quick fix
                    return CustomLicense(
                        str(licensing_info.license_id),
                        str(licensing_info.license_name),
                        str(licensing_info.extracted_text),
                    )

            return CustomLicense(license_expression.key, "", "")


def bump_license_exception(
    license_exception: LicenseSymbol, extracted_licensing_info: List[ExtractedLicensingInfo]
) -> LicenseAddition:
    if not get_spdx_licensing().validate(license_exception).invalid_symbols:
        return ListedLicenseException(license_exception.key, "", "")
    else:
        for licensing_info in extracted_licensing_info:
            if licensing_info.license_id == license_exception.key:
                # the fields are optional in ExtractedLicensingInfo, to prevent type errors we use a type conversion
                # to str as a quick fix
                return CustomLicenseAddition(
                    str(licensing_info.license_id),
                    str(licensing_info.license_name),
                    str(licensing_info.extracted_text),
                )
        return CustomLicenseAddition(license_exception.key, "", "")



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import Dict
from rdflib import DOAP, RDF, RDFS, XSD, BNode, Graph, Literal, URIRef

from spdx_tools.spdx.casing_tools import snake_case_to_camel_case
from spdx_tools.spdx.model import ExternalPackageRef, Package, PackageVerificationCode
from spdx_tools.spdx.model.package import CATEGORY_TO_EXTERNAL_PACKAGE_REF_TYPES
from spdx_tools.spdx.rdfschema.namespace import REFERENCE_NAMESPACE, SPDX_NAMESPACE
from spdx_tools.spdx.writer.rdf.checksum_writer import add_checksum_to_graph
from spdx_tools.spdx.writer.rdf.license_expression_writer import add_license_expression_or_none_or_no_assertion
from spdx_tools.spdx.writer.rdf.writer_utils import (
    add_datetime_to_graph,
    add_literal_or_no_assertion_or_none,
    add_namespace_to_spdx_id,
    add_optional_literal,
)


def add_package_to_graph(
    package: Package, graph: Graph, doc_namespace: str, external_doc_ref_to_namespace: Dict[str, str]
):
    package_resource = URIRef(add_namespace_to_spdx_id(package.spdx_id, doc_namespace, external_doc_ref_to_namespace))
    graph.add((package_resource, RDF.type, SPDX_NAMESPACE.Package))

    graph.add((package_resource, SPDX_NAMESPACE.name, Literal(package.name)))
    add_optional_literal(package.version, graph, package_resource, SPDX_NAMESPACE.versionInfo)
    add_optional_literal(package.file_name, graph, package_resource, SPDX_NAMESPACE.packageFileName)
    add_optional_literal(package.supplier, graph, package_resource, SPDX_NAMESPACE.supplier)
    add_optional_literal(package.originator, graph, package_resource, SPDX_NAMESPACE.originator)
    add_literal_or_no_assertion_or_none(
        package.download_location, graph, package_resource, SPDX_NAMESPACE.downloadLocation
    )
    graph.add((package_resource, SPDX_NAMESPACE.filesAnalyzed, Literal(package.files_analyzed, datatype=XSD.boolean)))
    add_package_verification_code_to_graph(package.verification_code, graph, package_resource)
    for checksum in package.checksums:
        add_checksum_to_graph(checksum, graph, package_resource)

    add_optional_literal(package.homepage, graph, package_resource, DOAP.homepage)
    add_optional_literal(package.source_info, graph, package_resource, SPDX_NAMESPACE.sourceInfo)
    add_license_expression_or_none_or_no_assertion(
        package.license_concluded, graph, package_resource, SPDX_NAMESPACE.licenseConcluded, doc_namespace
    )
    add_license_expression_or_none_or_no_assertion(
        package.license_info_from_files, graph, package_resource, SPDX_NAMESPACE.licenseInfoFromFiles, doc_namespace
    )
    add_license_expression_or_none_or_no_assertion(
        package.license_declared, graph, package_resource, SPDX_NAMESPACE.licenseDeclared, doc_namespace
    )
    add_optional_literal(package.license_comment, graph, package_resource, SPDX_NAMESPACE.licenseComments)
    add_optional_literal(package.copyright_text, graph, package_resource, SPDX_NAMESPACE.copyrightText)
    add_optional_literal(package.summary, graph, package_resource, SPDX_NAMESPACE.summary)
    add_optional_literal(package.description, graph, package_resource, SPDX_NAMESPACE.description)
    add_optional_literal(package.comment, graph, package_resource, RDFS.comment)
    for external_reference in package.external_references:
        add_external_package_ref_to_graph(external_reference, graph, package_resource, doc_namespace)
    for attribution_text in package.attribution_texts:
        add_optional_literal(attribution_text, graph, package_resource, SPDX_NAMESPACE.attributionText)
    if package.primary_package_purpose:
        graph.add(
            (
                package_resource,
                SPDX_NAMESPACE.primaryPackagePurpose,
                SPDX_NAMESPACE[f"purpose_{snake_case_to_camel_case(package.primary_package_purpose.name)}"],
            )
        )

    add_datetime_to_graph(package.release_date, graph, package_resource, SPDX_NAMESPACE.releaseDate)
    add_datetime_to_graph(package.built_date, graph, package_resource, SPDX_NAMESPACE.builtDate)
    add_datetime_to_graph(package.valid_until_date, graph, package_resource, SPDX_NAMESPACE.validUntilDate)


def add_package_verification_code_to_graph(
    package_verification_code: PackageVerificationCode, graph: Graph, package_node: URIRef
):
    if not package_verification_code:
        return
    package_verification_code_node = BNode()
    graph.add((package_verification_code_node, RDF.type, SPDX_NAMESPACE.PackageVerificationCode))
    graph.add(
        (
            package_verification_code_node,
            SPDX_NAMESPACE.packageVerificationCodeValue,
            Literal(package_verification_code.value),
        )
    )
    for excluded_file in package_verification_code.excluded_files:
        graph.add(
            (
                package_verification_code_node,
                SPDX_NAMESPACE.packageVerificationCodeExcludedFile,
                Literal(excluded_file),
            )
        )

    graph.add((package_node, SPDX_NAMESPACE.packageVerificationCode, package_verification_code_node))


def add_external_package_ref_to_graph(
    external_package_ref: ExternalPackageRef, graph: Graph, package_node: URIRef, doc_namespace: str
):
    external_package_ref_node = BNode()
    graph.add((external_package_ref_node, RDF.type, SPDX_NAMESPACE.ExternalRef))
    graph.add(
        (
            external_package_ref_node,
            SPDX_NAMESPACE.referenceCategory,
            SPDX_NAMESPACE[f"referenceCategory_{snake_case_to_camel_case(external_package_ref.category.name)}"],
        )
    )

    if external_package_ref.reference_type in CATEGORY_TO_EXTERNAL_PACKAGE_REF_TYPES[external_package_ref.category]:
        graph.add(
            (
                external_package_ref_node,
                SPDX_NAMESPACE.referenceType,
                REFERENCE_NAMESPACE[external_package_ref.reference_type],
            )
        )
    else:
        graph.add(
            (
                external_package_ref_node,
                SPDX_NAMESPACE.referenceType,
                URIRef(f"{doc_namespace}#{external_package_ref.reference_type}"),
            )
        )
    graph.add((external_package_ref_node, SPDX_NAMESPACE.referenceLocator, Literal(external_package_ref.locator)))
    if external_package_ref.comment:
        graph.add((external_package_ref_node, RDFS.comment, Literal(external_package_ref.comment)))

    graph.add((package_node, SPDX_NAMESPACE.externalRef, external_package_ref_node))



# Copyright (c) 2014 Ahmed H. Ismail
# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

import pytest

from spdx_tools.spdx.constants import DOCUMENT_SPDX_ID
from spdx_tools.spdx.model import Relationship, RelationshipType
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.tagvalue.parser import Parser
from tests.spdx.parser.tagvalue.test_creation_info_parser import DOCUMENT_STR


def test_parse_unknown_tag():
    parser = Parser()
    unknown_tag_str = "UnknownTag: This is an example for an unknown tag."

    with pytest.raises(SPDXParsingError, match="Unknown tag"):
        parser.parse(unknown_tag_str)


def test_building_contains_relationship():
    parser = Parser()
    document_str = "\n".join(
        [
            DOCUMENT_STR,
            "FileName: File without package",
            "SPDXID: SPDXRef-File",
            "FileChecksum: SHA1: d6a770ba38583ed4bb4525bd96e50461655d2759",
            "PackageName: Package with two files",
            "SPDXID: SPDXRef-Package-with-two-files",
            "PackageDownloadLocation: https://download.com",
            "FileName: File in package",
            "SPDXID: SPDXRef-File-in-Package",
            "FileChecksum: SHA1: d6a770ba38583ed4bb4525bd96e50461655d2759",
            "FileName: Second file in package",
            "SPDXID: SPDXRef-Second-File-in-Package",
            "FileChecksum: SHA1: d6a770ba38583ed4bb4525bd96e50461655d2759",
            "PackageName: Second package with file",
            "SPDXID: SPDXRef-Package-with-one-file",
            "PackageDownloadLocation: https://download.com",
            "FileName: File in package",
            "SPDXID: SPDXRef-File-in-different-Package",
            "FileChecksum: SHA1: d6a770ba38583ed4bb4525bd96e50461655d2759",
        ]
    )
    document = parser.parse(document_str)

    assert document.relationships == [
        Relationship("SPDXRef-Package-with-two-files", RelationshipType.CONTAINS, "SPDXRef-File-in-Package"),
        Relationship("SPDXRef-Package-with-two-files", RelationshipType.CONTAINS, "SPDXRef-Second-File-in-Package"),
        Relationship("SPDXRef-Package-with-one-file", RelationshipType.CONTAINS, "SPDXRef-File-in-different-Package"),
    ]


def test_build_contains_relationship_with_error():
    parser = Parser()
    file_spdx_ids = ["SPDXRef-File-in-Package", "SPDXRef-Second-File-in-Package"]
    document_str = "\n".join(
        [
            DOCUMENT_STR,
            "PackageName: Package with two files",
            "PackageDownloadLocation: https://download.com",
            "FileName: File in package",
            f"SPDXID: {file_spdx_ids[0]}",
            "FileChecksum: SHA1: d6a770ba38583ed4bb4525bd96e50461655d2759",
            "FileName: Second file in package",
            f"SPDXID: {file_spdx_ids[1]}",
            "FileChecksum: SHA1: d6a770ba38583ed4bb4525bd96e50461655d2759",
        ]
    )
    with pytest.raises(SPDXParsingError) as err:
        parser.parse(document_str)
    for file_spdx_id in file_spdx_ids:
        assert (
            f"Error while building contains relationship for file {file_spdx_id}, preceding package was not "
            "parsed successfully." in err.value.get_messages()
        )


def test_document_with_mixed_values():
    parser = Parser()
    document_str = "\n".join(
        [
            f"SPDXID:{DOCUMENT_SPDX_ID}",
            "FileName: File without package",
            "SPDXID: SPDXRef-File",
            "PackageDownloadLocation: https://download.com",
            "FileChecksum: SHA1: d6a770ba38583ed4bb4525bd96e50461655d2759",
        ]
    )

    with pytest.raises(SPDXParsingError) as err:
        parser.parse(document_str)

    assert err.value.get_messages() == [
        "Element Package is not the current element in scope, probably the expected "
        "tag to start the element (PackageName) is missing. Line: 4"
    ]



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from enum import Enum, auto

from beartype.typing import Optional, Union

from spdx_tools.common.typing.dataclass_with_properties import dataclass_with_properties
from spdx_tools.common.typing.type_checks import check_types_and_set_values
from spdx_tools.spdx.model import SpdxNoAssertion, SpdxNone


class RelationshipType(Enum):
    AMENDS = auto()
    ANCESTOR_OF = auto()
    BUILD_DEPENDENCY_OF = auto()
    BUILD_TOOL_OF = auto()
    CONTAINED_BY = auto()
    CONTAINS = auto()
    COPY_OF = auto()
    DATA_FILE_OF = auto()
    DEPENDENCY_MANIFEST_OF = auto()
    DEPENDENCY_OF = auto()
    DEPENDS_ON = auto()
    DESCENDANT_OF = auto()
    DESCRIBED_BY = auto()
    DESCRIBES = auto()
    DEV_DEPENDENCY_OF = auto()
    DEV_TOOL_OF = auto()
    DISTRIBUTION_ARTIFACT = auto()
    DOCUMENTATION_OF = auto()
    DYNAMIC_LINK = auto()
    EXAMPLE_OF = auto()
    EXPANDED_FROM_ARCHIVE = auto()
    FILE_ADDED = auto()
    FILE_DELETED = auto()
    FILE_MODIFIED = auto()
    GENERATED_FROM = auto()
    GENERATES = auto()
    HAS_PREREQUISITE = auto()
    METAFILE_OF = auto()
    OPTIONAL_COMPONENT_OF = auto()
    OPTIONAL_DEPENDENCY_OF = auto()
    OTHER = auto()
    PACKAGE_OF = auto()
    PATCH_APPLIED = auto()
    PATCH_FOR = auto()
    PREREQUISITE_FOR = auto()
    PROVIDED_DEPENDENCY_OF = auto()
    REQUIREMENT_DESCRIPTION_FOR = auto()
    RUNTIME_DEPENDENCY_OF = auto()
    SPECIFICATION_FOR = auto()
    STATIC_LINK = auto()
    TEST_CASE_OF = auto()
    TEST_DEPENDENCY_OF = auto()
    TEST_OF = auto()
    TEST_TOOL_OF = auto()
    VARIANT_OF = auto()


@dataclass_with_properties
class Relationship:
    spdx_element_id: str
    relationship_type: RelationshipType
    related_spdx_element_id: Union[str, SpdxNone, SpdxNoAssertion]
    comment: Optional[str] = None

    def __init__(
        self,
        spdx_element_id: str,
        relationship_type: RelationshipType,
        related_spdx_element_id: Union[str, SpdxNone, SpdxNoAssertion],
        comment: Optional[str] = None,
    ):
        check_types_and_set_values(self, locals())



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from unittest.mock import MagicMock, call, mock_open, patch

from spdx_tools.spdx.writer.tagvalue.snippet_writer import write_snippet
from tests.spdx.fixtures import snippet_fixture


def test_snippet_writer():
    snippet = snippet_fixture()

    mock: MagicMock = mock_open()
    with patch(f"{__name__}.open", mock, create=True):
        with open("foo", "w") as file:
            write_snippet(snippet, file)

    mock.assert_called_once_with("foo", "w")
    handle = mock()
    handle.write.assert_has_calls(
        [
            call("## Snippet Information\n"),
            call(f"SnippetSPDXID: {snippet.spdx_id}\n"),
            call(f"SnippetFromFileSPDXID: {snippet.file_spdx_id}\n"),
            call("SnippetByteRange: 1:2\n"),
            call("SnippetLineRange: 3:4\n"),
            call(f"SnippetLicenseConcluded: {snippet.license_concluded}\n"),
            call(f"LicenseInfoInSnippet: {snippet.license_info_in_snippet[0]}\n"),
            call(f"LicenseInfoInSnippet: {snippet.license_info_in_snippet[1]}\n"),
            call(f"LicenseInfoInSnippet: {snippet.license_info_in_snippet[2]}\n"),
            call(f"SnippetLicenseComments: {snippet.license_comment}\n"),
            call(f"SnippetCopyrightText: {snippet.copyright_text}\n"),
            call(f"SnippetComment: {snippet.comment}\n"),
            call(f"SnippetName: {snippet.name}\n"),
            call(f"SnippetAttributionText: {snippet.attribution_texts[0]}\n"),
        ]
    )
